#! /usr/bin/env python
import os
import os.path
import stat
import errno
import fuse
from fuse import Fuse
import sys
from time import time

from ThreadSafeDict import Dict
from PipelineMC import *

fuse.fuse_python_api = (0, 2)

accmode = os.O_RDONLY | os.O_WRONLY | os.O_RDWR

class Stat(fuse.Stat):
    def __init__(self):
        self.st_mode = 0
        self.st_ino = 0
        self.st_dev = 0
        self.st_nlink = 0
        self.st_uid = 0
        self.st_gid = 0
        self.st_size = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0
        self.st_blocks = 0
        self.st_blksize = 512
        self.st_rdev = 0
        return

class BackendDentry (object):

    dir_mode = stat.S_IFDIR | stat.S_IRWXU | \
        stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
    file_mode = stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR | \
        stat.S_IRGRP | stat.S_IROTH
    dir_access = ~0 & accmode
    file_access = ~(os.O_RDONLY | os.O_WRONLY | os.O_RDWR) & accmode
    xattr_prefix = 'user.edu.nrao.widar.'
    ENOATTR = errno.ENODATA

    def __init__(self, parent, name):
        self.client = parent.client
        self.parent = parent
        self.name = name
        self.root_node = parent.root_node
        if self.parent.be_path is not None:
            if self.name is not None:
                self.be_path = os.path.join(self.parent.be_path, self.name)
            else:
                self.be_path = self.parent.be_path
        else:
            self.be_path = self.name
        if self.be_path not in self.root_node.times:
            st = Stat()
            self.root_node.times[self.be_path] = st
            t = int(time())
            st.st_mtime = t
            st.st_atime = t
            st.st_ctime = t
        self.is_dir = False
        self.is_file = False
        return

    def __children(self):
        return {}
    children = property(lambda(self): self.__children())

    def mark_times(self, *fields):
        t = int(time())
        times = self.root_node.times[self.be_path]
        for f in fields:
            setattr(times, 'st_%s' % f, t)
        return

    def get_node(self, path):
        result = None
        if path == '' or path == '/':
            result = self
        else:
            children = self.children
            sep = path.find('/')
            if sep != -1:
                first = path[:sep]
                rest = path[sep + 1:]
            else:
                first = path
                rest = ''
            if first in children.keys():
                result = children[first]().get_node(rest)
        return result

class BackendDir (BackendDentry):

    mode = BackendDentry.dir_mode
    access = BackendDentry.dir_access

    def __init__(self, parent, name):
        super(BackendDir, self).__init__(parent, name)
        self.is_dir = True
        return

    def __attr(self):
        st = Stat()
        st.st_mode = self.mode
        st.st_nlink = 2 + self.num_child_dirs()
        return st
    attr = property(lambda(self): self.__attr())

    def num_child_dirs(self):
        result = 0
        for ch in self.children.values():
            if ch().is_dir:
                result += 1
        return result

class BackendFile (BackendDentry):

    mode = BackendDentry.file_mode
    access = BackendDentry.file_access

    def __init__(self, parent, name):
        super(BackendFile, self).__init__(parent, name)
        self.is_file = True
        return

    def __attr(self):
        st = Stat()
        st.st_mode = self.mode
        st.st_nlink = 1
        return st
    attr = property(lambda(self): self.__attr())

    def query(self):
        return ''

    def command(self, value):
        return -errno.EINVAL

    def read(self, size):
        result = self.query()
        return result[:min(len(result), size)]

    def write(self, buf):
        return self.command(buf)

class AliasedClient:
    def __init__(self, client, alias):
        self.client = client
        self.alias = alias
        return

    def send_msg(self, msg):
        return self.client.send_msg(self.alias, msg)

    def recv_msg(self):
        return self.client.recv_msg(self.alias)

    def mc_transaction(self, msg):
        return self.client.mc_transaction(self.alias, msg)

class Backend (BackendDir):

    def __init__(self, client, alias):
        self.client = AliasedClient(client, alias)
        self.be_path = ''
        self.new_config = None
        self.times = Dict()
        self.root_node = self
        self.times[''] = Stat()
        self.mark_times('ctime', 'mtime')
        return

    def __children(self):
        input_stage = PipelineClient.input_stage
        result = {'configurations': lambda: Configurations(self),
                  'flow' : lambda: FlowControl(self),
                  'src' : lambda: SrcControl(self),
                  'name' : lambda: RemoteName(self)}
        src = SrcControl(self).query()
        if src.startswith('file://'):
            result['rate'] = lambda: RateControl(self)
            result['live_ts'] = lambda: LiveTsControl(self)
        cfgs = [c() for c in
                Configurations(self).children.values()]
        active = [c.name for c in cfgs if
                  (isinstance(c, Config) and c.status[0].startswith('active'))]
        if len(active) == 1:
            result['active'] = lambda: ActiveLink(self, active[0])
        return result
    children = property(lambda(self): self.__children())

    def relative_path(path):
        while len(path) > 0 and path[0] == '/':
            path = path[1:]
        return path
    relative_path = staticmethod(relative_path)

    def get_listing(self, path):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        node.mark_times('atime')
        return node.children.keys()

    def getattr(self, path):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        if node is not None:
            result = node.attr
            st = self.root_node.times[node.be_path]
            result.st_atime = st.st_atime
            result.st_mtime = st.st_mtime
            result.st_ctime = st.st_ctime
        else:
            result = -errno.ENOENT
        return result

    def getxattr(self, path, name, size):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        if node is not None:
            try:
                name = name[len(self.xattr_prefix):]
                val =  node.xattr[name]
            except:
                result = -self.ENOATTR
            else:
                if size == 0:
                    result = len(val)
                else:
                    result = val
        else:
            result = -errno.ENOENT
        return result

    def listxattr(self, path, size):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        if node is not None:
            keys = [self.xattr_prefix + k for k in node.xattr.keys()]
            if size == 0:
                result = len(''.join(keys)) + len(keys)
            else:
                result = keys
        else:
            result = -errno.ENONENT
        return result

    def create(self, path, mode):
        path = Backend.relative_path(path)
        (parent, tail) = os.path.split(path)
        node = self.get_node(parent)
        if node is not None:
            result = node.create(tail)
            if result:
                new_node = self.get_node(path)
                if new_node is not None:
                    new_node.mark_times('mtime', 'ctime')
        else:
            result -errno.EACCES
        return result

    def open(self, path, flags):
        path = Backend.relative_path(path)
        node = self.get_node(path)
        if node is None:
            return -errno.ENOENT
        else:
            if flags & node.access != 0:
                return -errno.EACCES

    def read(self, path, size, offset):
        path = Backend.relative_path(path)
        if offset == 0:
            try:
                node = self.get_node(path)
                result = node.read(size)
                if len(result) > 0:
                    node.mark_times('atime')
            except:
                result = -errno.EINVAL
        else:
            result = None
        return result

    def write(self, path, buf, offset):
        path = Backend.relative_path(path)
        if offset == 0:
            try:
                node = self.get_node(path)
                result = node.write(buf)
                if result > 0:
                    node.mark_times('mtime')
            except:
                result = -errno.EINVAL
        else:
            result = None
        return result

    def readlink(self, path):
        path = Backend.relative_path(path)
        try:
            node = self.get_node(path)
            result = node.readlink()
            if len(result) > 0:
                node.mark_times('atime')
            return result
        except:
            return

    def rmdir(self, path):
        path = Backend.relative_path(path)
        try:
            node = self.get_node(path)
            if node.is_dir:
                result = node.rmdir()
                if result:
                    self.root_node.times.acquire()
                    removed = [n for n in self.root_node.times.keys() if
                               n.startswith(path)]
                    for r in removed:
                        del self.root_node.times[r]
                    self.root_node.times.release()
            else:
                result = -errno.ENOTDIR
        except:
            result = -errno.EPERM
        return result

class ActiveLink (BackendFile):

    mode = stat.S_IFLNK | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH

    def __init__(self, parent, active):
        super(ActiveLink, self).__init__(parent, None)
        self.linkval = os.path.join('configurations', active)
        return

    def __attr(self):
        st = Stat()
        st.st_mode = self.mode
        st.st_size = len(self.linkval)
        return st
    attr = property(lambda(self): self.__attr())

    def readlink(self):
        return self.linkval

class Configurations (BackendDir):

    def __init__(self, parent):
        super(Configurations, self).__init__(parent, None)
        return

    def __children(self):
        msg = ConfigListMsg(intent=MsgIntent.query)
        try:
            configs = self.client.mc_transaction(msg).configs
        except:
            configs = None
        result = {}
        if configs is not None:
            status = {}
            def add_status(cfgs, st):
                for c in cfgs:
                    if c not in status:
                        status[c] = []
                    status[c].append(st)
            add_status(configs.active, 'active')
            add_status(configs.active_retired, 'active-retired')
            add_status(configs.inactive, 'inactive')
            if len(configs.pending) > 0:
                for p in configs.pending:
                    (action, cfg) = p.split()
                    if cfg not in status:
                        status[cfg] = []
                    status[cfg].append('pending-%s' % action)
                result['pending'] = \
                    lambda: ConfigActions(self, configs.pending)
            def config(n, s):
                return lambda: Config(self, n, s)
            for (cfg, st) in status.items():
                result[cfg] = config(cfg, st)
        if self.parent.new_config is not None:
            result[self.parent.new_config] = \
                lambda: ConfigNew(self, self.parent.new_config)
        return result
    children = property(lambda(self): self.__children())

    def create(self, config):
        assert config not in self.children.keys()
        if self.parent.new_config is None:
            self.parent.new_config = config
        else:
            return -errno.EACCES
        return

class ConfigActions (BackendFile):

    mode = stat.S_IFREG | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH

    def __init__(self, parent, actions):
        super(ConfigActions, self).__init__(parent, None)
        self.actions = actions
        return

    def query(self):
        return '\n'.join(self.actions) + '\n'

class ConfigNew (BackendFile):

    def __init__(self, parent, name):
        super(ConfigNew, self).__init__(parent, None)
        self.config_name = name
        return

    def command(self, value):
        msg = ConfigCreateMsg(name=self.config_name, desc=value,
                              intent=MsgIntent.command)
        try:
            self.client.mc_transaction(msg)
            result = len(value)
        except:
            result = -errno.EINVAL
        self.parent.parent.new_config = None
        return result

class Config (BackendDir):

    def __init__(self, parent, name, status):
        super(Config, self).__init__(parent, name)
        self.status = status
        return

    def __children(self):
        result = {'status' : lambda: ConfigStatus(self, self.status),
                  'pipelines' : lambda: PipelineDir(self)}
        return result
    children = property(lambda(self): self.__children())

    def rmdir(self):
        msg = ConfigDestroyMsg(name=self.name, intent=MsgIntent.command)
        try:
            self.client.mc_transaction(msg)
            result = 0
        except:
            result = -errno.EINVAL # TODO: correct error value?
        return result

class PipelineDir (BackendDir):

    def __init__(self, parent):
        super(PipelineDir, self).__init__(parent, None)
        return

    def __children(self):
        msg = PipelineListMsg(config=self.be_path, intent=MsgIntent.query)
        try:
            pipelines = self.client.mc_transaction(msg).pipelines
        except:
            pipelines = []
        result = {}
        # the following fn is required to bind the pipeline name in
        # the lambda expression
        def pipeline(name):
            return lambda: Pipeline(self, name)
        for p in pipelines:
            result[p] = pipeline(p)
        return result
    children = property(lambda(self): self.__children())

class ConfigStatus (BackendFile):

    def __init__(self, parent, status):
        super(ConfigStatus, self).__init__(parent, None)
        self.status = status
        return

    def query(self):
        return '\n'.join(self.status) + '\n'

    def command(self, value):
        cmd = value.strip()
        if cmd == 'activate':
            msg = ConfigActivateMsg(name=self.parent.name,
                                    intent=MsgIntent.command)
        elif cmd == 'destroy':
            msg = ConfigDestroyMsg(name=self.parent.name,
                                   intent=MsgIntent.command)
        else:
            msg = None
        result = -errno.EINVAL
        if msg is not None:
            try:
                self.client.mc_transaction(msg)
                result = len(value)
            except:
                pass
        return result

class Pipeline (BackendDir):

    def __init__(self, parent, name):
        super(Pipeline, self).__init__(parent, name)
        return

    def __children(self):
        msg = ElementListMsg(pipeline=self.be_path, intent=MsgIntent.query)
        try:
            elements = self.client.mc_transaction(msg).elements
        except:
            elements = []
        result = {}
        def element(name):
            return lambda: Element(self, name)
        for e in elements:
            result[e] = element(e)
        return result
    children = property(lambda(self): self.__children())

class Element (BackendDir):

    def __init__(self, parent, name):
        super(Element, self).__init__(parent, name)
        return

    def __children(self):
        result = {'properties': lambda: PropertyDir(self),
                  'pads': lambda: PadDir(self)}
        return result
    children = property(lambda(self):self.__children())

class PropertyDir (BackendDir):

    def __init__(self, parent):
        super(PropertyDir, self).__init__(parent, None)
        return

    def __children(self):
        msg = PropertyListMsg(element=self.be_path, intent=MsgIntent.query)
        try:
            properties = self.client.mc_transaction(msg).properties
        except:
            properties = []
        result = {}
        def prop(name, type):
            return lambda: Property(self, name, type)
        for p in properties:
            langle = p.find('<')
            assert langle != -1
            rangle = p.rfind('>')
            assert rangle != -1
            name = p[:langle]
            valtype = p[langle + 1:rangle]
            result[name] = prop(name, valtype)
        return result
    children = property(lambda(self): self.__children())

class PadDir (BackendDir):

    def __init__(self, parent):
        super(PadDir, self).__init__(parent, None)
        return

    def __children(self):
        msg = PadListMsg(element=self.be_path, intent=MsgIntent.query)
        try:
            pads = self.client.mc_transaction(msg).pads
        except:
            pads = []
        result = {}
        def pad(name):
            return lambda: Pad(self, name)
        for p in pads:
            result[p] = pad(p)
        return result
    children = property(lambda(self): self.__children())

class Pad (BackendDir):

    def __init__(self, parent, name):
        super(Pad, self).__init__(parent, name)
        return

    def __children(self):
        result = {'peer': lambda: PadPeer(self),
                  'caps': lambda: PadCaps(self)}
        return result
    children = property(lambda(self): self.__children())

class PadPeer (BackendFile):

    mode = stat.S_IFREG | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH

    def __init__(self, parent):
        super(PadPeer, self).__init__(parent, None)
        return

    def query(self):
        msg = PadPeerMsg(pad=self.be_path, intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).peer
        except:
            result = ''
        return result + '\n'

class PadCaps (BackendFile):

    mode = stat.S_IFREG | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH

    def __init__(self, parent):
        super(PadCaps, self).__init__(parent, None)
        return

    def query(self):
        msg = PadCapsMsg(pad=self.be_path, intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).caps
        except:
            result = ''
        return result + '\n'

class FlowControl (BackendFile):

    def __init__(self, parent):
        super(FlowControl, self).__init__(parent, None)
        return

    def query(self):
        msg = FlowMsg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).state.nick()
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = FlowMsg(intent=MsgIntent.command, state=Flow(value.strip()))
            self.client.mc_transaction(msg)
            result = len(value)
        except:
            result = -errno.EINVAL
        return result

class SrcControl (BackendFile):

    def __init__(self, parent):
        super(SrcControl, self).__init__(parent, None)
        return

    def query(self):
        msg = SrcMsg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).name
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = SrcMsg(intent=MsgIntent.command, name=value.strip())
            self.client.mc_transaction(msg)
            result = len(value)
        except:
            result = -errno.EINVAL
        return result

class RateControl (BackendFile):

    def __init__(self, parent):
        super(RateControl, self).__init__(parent, None)
        return

    def query(self):
        msg = RateMsg(intent=MsgIntent.query)
        try:
            result = str(self.client.mc_transaction(msg).rate)
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = RateMsg(intent=MsgIntent.command, rate=int(value.strip()))
            self.client.mc_transaction(msg)
            result = len(value)
        except:
            result = -errno.EINVAL
        return result

class LiveTsControl (BackendFile):

    def __init__(self, parent):
        super(LiveTsControl, self).__init__(parent, None)
        return

    def query(self):
        msg = LiveTsMsg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).state.nick()
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = LiveTsMsg(intent=MsgIntent.command,
                            state=LiveTs(value.strip()))
            self.client.mc_transaction(msg)
            result = len(value)
        except:
            result = -errno.EINVAL
        return result

class RemoteName (BackendFile):

    mode = stat.S_IFREG | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH

    def __init__(self, parent):
        super(RemoteName, self).__init__(parent, None)
        return

    def query(self):
        msg = NameMsg(intent=MsgIntent.query)
        try:
            result = self.client.mc_transaction(msg).name
        except:
            result = ''
        return result + '\n'

class Property (BackendFile):

    def __init__(self, parent, name, type):
        super(Property, self).__init__(parent, name)
        self.type = type
        return

    def __xattr(self):
        return {'gtype': self.type}
    xattr = property(lambda (self): self.__xattr())

    def query(self):
        msg = PropertyMsg(prop=self.be_path, valtype=self.type,
                          intent=MsgIntent.query)
        try:
            result = str(self.client.mc_transaction(msg).value)
        except:
            result = ''
        return result + '\n'

    def command(self, value):
        try:
            msg = PropertyMsg(prop=self.be_path, valtype=self.type,
                              value=value.strip(), intent=MsgIntent.command)
            self.client.mc_transaction(msg)
            result = len(value)
        except:
            result = -errno.EINVAL
        return result

class PipelineFS (Fuse):

    def main(self):
        self.pipeline_state = []
        statefile_path = os.path.abspath(
            os.path.join(self.fuse_args.mountpoint, '.state'))
        try:
            statefile = open(statefile_path, 'r')
            for line in statefile:
                p = line.split()
                if len(p) == 2:
                    self.pipeline_state.append(tuple(p))
            statefile.close()
        except:
            pass
        result = Fuse.main(self)
        statefile = open(statefile_path, 'w')
        for t in self.trees:
            (be, addr) = self.trees[t]
            statefile.write('%s\t%s\n' % (addr, t))
        statefile.close()
        return result
    
    def fsinit(self):
        self.client = PipelineClient()
        self.trees = {}
        self.new_trees = []
        self.times = Dict()
        self.times[''] = Stat()
        self.mark_times('', 'ctime', 'mtime')
        for (addr, alias) in self.pipeline_state:
            self.new_trees.append(alias)
            self.times[alias] = Stat()
            rc = self.link_pipeline(alias, addr, False)
        return

    def link_pipeline(self, alias, addr, allow_new=True):
        self.new_trees.remove(alias)
        del self.times[alias]
        name_parts = addr.split(':')
        rc = False
        if len(name_parts) == 1:
            rc = self.client.add_pipeline(alias=alias,
                                          name=name_parts[0],
                                          allow_new=allow_new)
        elif len(name_parts) == 2:
            rc = self.client.add_pipeline(alias=alias,
                                          host=name_parts[0],
                                          port=int(name_parts[1]),
                                          allow_new=allow_new)
        elif len(name_parts) == 3:
            rc = self.client.add_pipeline(alias=alias,
                                          host=name_parts[0],
                                          port=int(name_parts[1]),
                                          name=name_parts[2],
                                          allow_new=allow_new)
        if rc:
            self.trees[alias] = (Backend(alias=alias, client=self.client),
                                 addr)
            result = 0
        else:
            result = -errno.ENXIO # TODO: return a better error code?
        return result
    
    def mark_times(self, path, *fields):
        t = int(time())
        try:
            times = self.times[path]
        except KeyError:
            times = Stat()
            self.times[path] = times
        for f in fields:
            setattr(times, 'st_%s' % f, t)
        return

    def split_pipeline_path(self, path):
        if path[0] == '/':
            path = path[1:]
        slash = path.find('/')
        if slash != -1:
            result = (path[:slash], path[slash:])
        else:
            result = (path, '')
        return result

    def getattr(self, path):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].getattr(path)
        except KeyError:
            if path == '':
                st = Stat()
                if alias == '':
                    st.st_mode = stat.S_IFDIR | stat.S_IRWXU | \
                        stat.S_IRGRP | stat.S_IROTH | \
                        stat.S_IXGRP | stat.S_IXOTH
                    st.st_nlink = 2 + len(self.trees.keys())
                elif alias in self.new_trees:
                    st.st_mode = stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR | \
                        stat.S_IRGRP | stat.S_IROTH
                    st.st_nlink = 1
                else:
                    st = -errno.ENOENT
                if st != -errno.ENOENT:
                    st.st_atime = self.times[alias].st_atime
                    st.st_mtime = self.times[alias].st_mtime
                    st.st_ctime = self.times[alias].st_ctime
                result = st
            else:
                result = None
        return result

    def readdir(self, path, offset):
        (alias, path) = self.split_pipeline_path(path)
        if alias != '':
            try:
                tree = self.trees[alias][0]
            except KeyError:
                pass
            else:
                listing = tree.get_listing(path)
                for e in listing:
                    yield fuse.Direntry(e)
        else:
            for t in self.trees:
                self.trees[t][0].mark_times('atime')
                yield fuse.Direntry(t)
            for t in self.new_trees:
                self.mark_times(t, 'atime')
                yield fuse.Direntry(t)
        return

    def open(self, path, flags):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].open(path, flags)
        except KeyError:
            if path == '' and alias in self.new_trees:
                result = 0
            else:
                result = -errno.ENOENT
        return result

    def create(self, path, mode, info):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].create(path, mode)
        except KeyError:
            if path == '':
                self.mark_times(alias, 'ctime', 'mtime')
                self.new_trees.append(alias)
                result = 0
            else:
                result = -errno.EFAULT
        return result

    def mknod(self, path, mode, dev):
        if mode & stat.S_IFREG:
            result = self.create(path, mode, '')
        else:
            result = None
        return result

    def read(self, path, size, offset):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].read(path, size, offset)
        except KeyError:
            if path == '' and alias in self.new_trees:
                self.mark_times(alias, 'atime')
                result = ''
            else:
                result = -errno.EINVAL
        return result

    def write(self, path, buf, offset):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].write(path, buf, offset)
        except KeyError:
            if path == '' and alias in self.new_trees:
                result = self.link_pipeline(alias, buf.strip())
                if result == 0:
                    result = len(buf)
            else:
                result = -errno.EINVAL
        return result

    def truncate(self, path, len):
        return

    def rmdir(self, path):
        (alias, path) = self.split_pipeline_path(path)
        try:
            tree = self.trees[alias][0]
        except KeyError:
            result = -errno.EFAULT
        else:
            if path != '' and path != '/':
                result = self.trees[alias][0].rmdir(path)
            else:
                del self.trees[alias]
                self.client.remove_pipeline(alias)
                result = 0
        return result

    def readlink(self, path):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].readlink(path)
        except KeyError:
            result = None
        return result

    def getxattr(self, path, name, size):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].getxattr(path, name, size)
        except KeyError:
            result = None
        return result

    def listxattr(self, path, size):
        (alias, path) = self.split_pipeline_path(path)
        try:
            result = self.trees[alias][0].listxattr(path, size)
        except KeyError:
            result = None
        return result

def main():

    description="Filesystem reflecting WCBE compute node pipeline."

    server = PipelineFS(description=description)
    server.parse(errex=1)
    optdict = server.fuse_args.optdict
    if 'uid' not in optdict:
        optdict['uid'] = os.geteuid()
    if 'gid' not in optdict:
        optdict['gid'] = os.getegid()
    server.fuse_args.optlist.add('direct_io')
    server.fuse_args.optlist.add('nonempty')

    #if server.fuse_args.mount_expected():
    #    os.chdir(server.root)

    server.main()

if __name__ == '__main__':
    main()
