import socket
import stat
import os
import struct
import threading
import select
import errno
from tempfile import mkstemp
from time import time, sleep
from ThreadSafeDict import Dict

from Enum import Enum

class DumpFormat (object):
    __metaclass__ = Enum
    __enums = {'ascii':0, 'binary':1, 'none':2}

class Flow (object):
    __metaclass__ = Enum
    __enums = {'off':0, 'on':1}

class LiveTs (object):
    __metaclass__ = Enum
    __enums = {'false':0, 'true':1}

class MsgType (object):
    __metaclass__ = Enum
    __enums = {'config_create':0, 'config_activate':1, 'config_destroy':2,
               'config_list':3, 'src':4, 'live_ts':5, 'framedump':6,
               'flow':7, 'rate':8, 'name':9, 'quit':10,
               'pipeline_list':11, 'element_list':12, 'property_list':13,
               'property':14, 'pad_list':15, 'pad_caps':16, 'pad_peer':17}

class MsgDirection (object):
    __metaclass__ = Enum
    __enums = {'request':0, 'response':1}

class MsgIntent (object):
    __metaclass__ = Enum
    __enums = {'command':0, 'query':1}

def str_trunc(str):
    null = str.find('\0')
    if null == -1:
        result = str
    else:
        result = str[:null]
    return result

class CbeMsg (object):

    format_str = ''
    type = ''

    def __init__(self, direction, intent, id):
        self.fields = {}
        self.fields['intent'] = intent
        self.fields['direction'] = direction
        self.id = id
        return

    intent = property(lambda (self):self.fields['intent'])
    direction = property(lambda (self):self.fields['direction'])

    def msg_classes():
        return (ConfigCreateMsg, ConfigActivateMsg, ConfigDestroyMsg,
                ConfigListMsg, SrcMsg, LiveTsMsg, FramedumpMsg,
                FlowMsg, RateMsg, NameMsg, QuitMsg, PipelineListMsg,
                ElementListMsg, PropertyListMsg, PropertyMsg,
                PadListMsg, PadCapsMsg, PadPeerMsg)
    msg_classes = staticmethod(msg_classes)

    def class_lookup(code):
        klass = [k for k in CbeMsg.msg_classes() if int(k.type) == code]
        if len(klass) != 1:
            raise IndexError()
        return klass[0]
    class_lookup = staticmethod(class_lookup)

    def pack(self):
        (fmt, args) = self.pack_args()
        return struct.pack('iiiI' + fmt, self.type,
                           self.direction, self.intent, self.id, *args)

    def unpack(str):
        size = len(str) - struct.calcsize('iiiI')
        (type, dir, intent, id, msg) = struct.unpack('iiiI%ds' % size, str)
        klass = CbeMsg.class_lookup(type)
        fields = (dir, intent, id) + klass.unpack(msg)
        return klass(*klass.convert_fields(*fields))
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id):
        return (MsgDirection(direction), MsgIntent(intent), id)
    convert_fields = staticmethod(convert_fields)

class StrMsg (CbeMsg):

    format_str = "256s"

    def __init__(self, intent, id=0, name='', direction=MsgDirection.request):
        super(StrMsg, self).__init__(direction, intent, id)
        null = name.find('\0')
        self.fields['name'] = str_trunc(name)
        return

    name = property(lambda (self):self.fields['name'])

    def pack_args(self):
        return (self.format_str, (self.name,))

    def unpack(str):
        return struct.unpack(StrMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, name):
        return (MsgIntent(intent), id, name, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class NameMsg (StrMsg):
    type = MsgType.name

class SrcMsg (StrMsg):
    type = MsgType.src

class ConfigCreateMsg (CbeMsg):

    type = MsgType.config_create
    format_str = '256sI%ds'

    def __init__(self, name, desc, intent=MsgIntent.command, id=0,
                 direction=MsgDirection.request):
        super(ConfigCreateMsg, self).__init__(direction, intent, id)
        self.fields['name'] = str_trunc(name)
        self.fields['desc'] = desc
        return

    name = property(lambda (self):self.fields['name'])
    desc = property(lambda (self):self.fields['desc'])

    def pack_args(self):
        desc_len = len(self.desc) + 1
        return (self.format_str % desc_len, (self.name, desc_len, self.desc))

    def unpack(str):
        size = len(str) - struct.calcsize(ConfigCreateMsg.format_str[:-3])
        return struct.unpack(ConfigCreateMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, name, desc_len, desc):
        return (name, desc, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class ConfigActivateMsg (StrMsg):
    type = MsgType.config_activate

class ConfigDestroyMsg (StrMsg):
    type = MsgType.config_destroy

class ConfigList:

    def __init__(self):
        self.active = ()
        self.active_retired = ()
        self.inactive = ()
        self.pending = ()
        return

    def encode(self):
        result = []
        for c in self.active:
            result.append('<%s>', c)
        for c in self.active_retired:
            result.append('[%s]', c)
        for c in self.inactive:
            result.append('(%s)', c)
        for c in self.pending:
            if c.startswith('create'):
                result.append('*%s*', c[len('create') + 1:])
            elif c.startswith('activate'):
                result.append('#%s#', c[len('activate') + 1:])
            elif c.startswith('destroy'):
                result.append('@%s@', c[len('destroy') + 1:])
        fmt = ''
        for cfg in result:
            fmt += '%ds' % (len(cfg) + 1)
        fmt += 'c'
        fields = tuple(result) + ('\0',)
        return struct.pack(fmt, *fields)

    def decode(self, configs_len, configs_str):
        configs = []
        if configs_len > 0:
            null = configs_str.find('\0')
            while null != 0:
                configs.append(configs_str[:null])
                configs_str = configs_str[null + 1:]
                null = configs_str.find('\0')
        def filter(chars):
            return [c[1:-1] for c in configs if c[0] in chars]
        self.active = tuple(filter('<'))
        self.active_retired = tuple(filter('['))
        self.inactive = tuple(filter('('))
        pending = []
        for c in configs:
            if c[0] == '*':
                pending.append('create %s' % c[1:-1])
            elif c[0] == '#':
                pending.append('activate %s' % c[1:-1])
            elif c[0] == '@':
                pending.append('destroy %s' % c[1:-1])
        self.pending = tuple(pending)
        return

class ConfigListMsg (CbeMsg):

    format_str = 'I%ds'
    type = MsgType.config_list

    def __init__(self, configs=ConfigList(),
                 intent=MsgIntent.query, id=0, direction=MsgDirection.request):
        super(ConfigListMsg, self).__init__(direction, intent, id)
        self.fields['configs'] = configs
        return

    configs = property(lambda (self):self.fields['configs'])

    def pack_args(self):
        result = self.fields['configs'].encode()
        return (self.format_str % len(result), (len(result), result))

    def unpack(str):
        size = len(str) - struct.calcsize(ConfigListMsg.format_str[:-3])
        return struct.unpack(ConfigListMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, configs_len, configs_str):
        configs = ConfigList()
        configs.decode(configs_len, configs_str)
        return (configs, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class LiveTsMsg (CbeMsg):

    format_str = "i"
    type = MsgType.live_ts

    def __init__(self, intent, id=0, state=LiveTs.false,
                 direction=MsgDirection.request):
        super(LiveTsMsg, self).__init__(direction, intent, id)
        self.fields['state'] = state
        return

    state = property(lambda (self):self.fields['state'])

    def pack_args(self):
        return (self.format_str, (self.state,))

    def unpack(str):
        return struct.unpack(LiveTsMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, state):
        return (MsgIntent(intent), id, LiveTs(state), MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class FramedumpMsg (CbeMsg):

    format_str = "i"
    type = MsgType.framedump

    def __init__(self, intent, id=0, format=DumpFormat.none,
                 direction=MsgDirection.request):
        super(FramedumpMsg, self).__init__(direction, intent, id)
        self.fields['format'] = format
        return

    format = property(lambda (self):self.fields['format'])

    def pack_args(self):
        return (self.format_str, (self.format,))

    def unpack(str):
        return struct.unpack(FramedumpMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, format):
        return (MsgIntent(intent), id, DumpFormat(format),
                MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class FlowMsg (CbeMsg):

    format_str = "i"
    type = MsgType.flow

    def __init__(self, intent, id=0, state=Flow.off,
                 direction=MsgDirection.request):
        super(FlowMsg, self).__init__(direction, intent, id)
        self.fields['state'] = state
        return

    state = property(lambda (self):self.fields['state'])

    def pack_args(self):
        return (self.format_str, (self.state,))

    def unpack(str):
        return struct.unpack(FlowMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, state):
        return (MsgIntent(intent), id, Flow(state), MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class RateMsg (CbeMsg):

    format_str = "I"
    type = MsgType.rate

    def __init__(self, intent, id=0, rate=0,
                 direction=MsgDirection.request):
        super(RateMsg, self).__init__(direction, intent, id)
        self.fields['rate'] = rate
        return

    rate = property(lambda (self):self.fields['rate'])

    def pack_args(self):
        return (self.format_str, (self.rate,))

    def unpack(str):
        return struct.unpack(RateMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, rate):
        return (MsgIntent(intent), id, rate, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class QuitMsg (CbeMsg):

    format_str = "I"
    type = MsgType.quit

    def __init__(self, id=0, intent=MsgIntent.command,
                 direction=MsgDirection.request):
        super(QuitMsg, self).__init__(direction, intent, id)
        return

    def pack_args(self):
        return (self.format_str, (0,))

    def unpack(str):
        return struct.unpack(QuitMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, reserved):
        return (id, MsgIntent(intent), MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)


class StringListMsg (CbeMsg):

    format_str = '256sI%ds'

    def __init__(self, intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(StringListMsg, self).__init__(direction, intent, id)
        return

    def pack_args(self, name_field, list_field):
        fmt = ''
        for p in self.fields[list_field]:
            fmt += '%ds' % (len(p) + 1)
        fmt += 'c'
        fields = self.fields[list_field] + ('\0',)
        result = struct.pack(fmt, *fields)
        return (self.format_str % len(result),
                (self.fields[name_field], len(result), result))

    def unpack(str):
        size = len(str) - struct.calcsize(StringListMsg.format_str[:-3])
        return struct.unpack(StringListMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, name, list_len, list_str):
        list = []
        if list_len > 0:
            null = list_str.find('\0')
            while null != 0:
                list.append(list_str[:null])
                list_str = list_str[null + 1:]
                null = list_str.find('\0')
        return (name, tuple(list), MsgIntent(intent), id,
                MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class PipelineListMsg (StringListMsg):

    type = MsgType.pipeline_list

    def __init__(self, config='', pipelines=(), intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PipelineListMsg, self).__init__(intent, id, direction)
        self.fields['config'] = str_trunc(config)
        self.fields['pipelines'] = pipelines
        return

    config = property(lambda(self):self.fields['config'])
    pipelines = property(lambda (self):self.fields['pipelines'])

    def pack_args(self):
        return super(PipelineListMsg, self).pack_args('config', 'pipelines')

class ElementListMsg (StringListMsg):

    type = MsgType.element_list

    def __init__(self, pipeline='', elements=(), intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(ElementListMsg, self).__init__(intent, id, direction)
        self.fields['pipeline'] = str_trunc(pipeline)
        self.fields['elements'] = elements
        return

    pipeline = property(lambda(self):self.fields['pipeline'])
    elements = property(lambda (self):self.fields['elements'])

    def pack_args(self):
        return super(ElementListMsg, self).pack_args('pipeline', 'elements')

class PropertyListMsg (StringListMsg):

    type = MsgType.property_list

    def __init__(self, element='', properties=(), intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PropertyListMsg, self).__init__(intent, id, direction)
        self.fields['element'] = str_trunc(element)
        self.fields['properties'] = properties
        return

    element= property(lambda(self):self.fields['element'])
    properties = property(lambda (self):self.fields['properties'])

    def pack_args(self):
        return super(PropertyListMsg, self).pack_args('element', 'properties')

class PadListMsg (StringListMsg):

    type = MsgType.pad_list

    def __init__(self, element='', pads=(), intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PadListMsg, self).__init__(intent, id, direction)
        self.fields['element'] = str_trunc(element)
        self.fields['pads'] = pads
        return

    element= property(lambda(self):self.fields['element'])
    pads = property(lambda (self):self.fields['pads'])

    def pack_args(self):
        return super(PadListMsg, self).pack_args('element', 'pads')

class PropertyMsg (CbeMsg):

    type = MsgType.property

    def __init__(self, prop='', value=None, valtype='',
                 intent=MsgIntent.query, id=0, direction=MsgDirection.request):
        super(PropertyMsg, self).__init__(direction, intent, id)
        self.fields['prop'] = str_trunc(prop)
        self.valtype = valtype
        format_str = '256s256sI'
        if value is None:
            self.fields['value'] = None
            self.format_str = format_str
            self.value_len = 0
        else:
            if valtype == 'gchar':
                self.fields['value'] = ord(str(value)[0])
                self.value_len = struct.calcsize('b')
                self.format_str = format_str + "b"
            elif valtype == 'guchar':
                self.fields['value'] = ord(str(value)[0])
                self.value_len = struct.calcsize('B')
                self.format_str = format_str + "B"
            elif valtype == 'gboolean':
                self.fields['value'] = int(value)
                self.value_len = struct.calcsize('i')
                self.format_str = format_str + 'i'
            elif valtype == 'gint':
                self.fields['value'] = int(value)
                self.value_len = struct.calcsize('i')
                self.format_str = format_str + 'i'
            elif valtype == 'guint':
                self.fields['value'] = long(value)
                self.value_len = struct.calcsize('I')
                self.format_str = format_str + 'I'
            elif valtype == 'glong':
                self.fields['value'] = int(value)
                self.value_len = struct.calcsize('l')
                self.format_str = format_str + 'l'
            elif valtype == 'gulong':
                self.fields['value'] = long(value)
                self.value_len = struct.calcsize('L')
                self.format_str = format_str + 'L'
            elif valtype == 'gint64':
                self.fields['value'] = long(value)
                self.value_len = struct.calcsize('q')
                self.format_str = format_str + 'q'
            elif valtype == 'guint64':
                self.fields['value'] = long(value)
                self.value_len = struct.calcsize('Q')
                self.format_str = format_str + 'Q'
            elif valtype == 'gfloat':
                self.fields['value'] = float(value)
                self.value_len = struct.calcsize('f')
                self.format_str = format_str + 'f'
            elif valtype == 'gdouble':
                self.fields['value'] = float(value)
                self.value_len = struct.calcsize('d')
                self.format_str = format_str + 'd'
            elif valtype == 'WcbeFramedumpFormat':
                self.fields['value'] = int(DumpFormat(value))
                self.value_len = struct.calcsize('i')
                self.format_str = format_str + 'i'
            else: # either valtype == 'gchararray' or something else
                self.fields['value'] = str(value)
                self.value_len = len(self.fields['value']) + 1
                self.format_str = format_str + ("%ds" % self.value_len)
        return

    prop = property(lambda(self):self.fields['prop'])
    value = property(lambda(self):self.fields['value'])

    def pack_args(self):
        if self.value is None:
            result = (self.format_str, (self.prop, '', 0))
        else:
            result = (self.format_str,
                      (self.prop, self.valtype, self.value_len, self.value))
        return result

    def unpack(str):
        size = len(str) - struct.calcsize('256s256sI')
        return struct.unpack('256s256sI%ds' % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, property, valtype,
                       value_len, value_str):
        valtype = str_trunc(valtype)
        if valtype == 'gchararray':
            value = value_str
        elif valtype == 'gchar':
            (value,) = struct.unpack('b', value_str)
        elif valtype == 'guchar':
            (value,) = struct.unpack('B', value_str)
        elif valtype == 'gboolean':
            (bool,) = struct.unpack('i', value_str)
            if bool == 0:
                value = False
            else:
                value = True
        elif valtype == 'gint':
            (value,) = struct.unpack('i', value_str)
        elif valtype == 'guint':
            (value,) = struct.unpack('I', value_str)
        elif valtype == 'glong':
            (value,) = struct.unpack('l', value_str)
        elif valtype == 'gulong':
            (value,) = struct.unpack('L', value_str)
        elif valtype == 'gint64':
            (value,) = struct.unpack('q', value_str)
        elif valtype == 'guint64':
            (value,) = struct.unpack('Q', value_str)
        elif valtype == 'gfloat':
            (value,) = struct.unpack('f', value_str)
        elif valtype == 'gdouble':
            (value,) = struct.unpack('d', value_str)
        elif valtype == 'WcbeFramedumpFormat':
            (enum,) = struct.unpack('i', value_str)
            if enum == 0:
                value = 'ascii'
            elif enum == 1:
                value = 'binary'
            else:
                value = 'none'
        else:
            value = value_str # FIXME: convert enums and flags
        return (property, value, valtype, MsgIntent(intent), id,
                MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class PadCapsMsg (CbeMsg):

    format_str = "256s256s"
    type = MsgType.pad_caps

    def __init__(self, pad='', caps='', intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PadCapsMsg, self).__init__(direction, intent, id)
        self.fields['pad'] = str_trunc(pad)
        self.fields['caps'] = str_trunc(caps)
        return

    pad = property(lambda (self):self.fields['pad'])
    caps = property(lambda (self):self.fields['caps'])

    def pack_args(self):
        return (self.format_str, (self.pad, self.caps))

    def unpack(str):
        return struct.unpack(PadCapsMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, pad, caps):
        return (pad, caps, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class PadPeerMsg (CbeMsg):

    format_str = "256s256s"
    type = MsgType.pad_peer

    def __init__(self, pad='', peer='', intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(PadPeerMsg, self).__init__(direction, intent, id)
        self.fields['pad'] = str_trunc(pad)
        self.fields['peer'] = str_trunc(peer)
        return

    pad = property(lambda (self):self.fields['pad'])
    peer = property(lambda (self):self.fields['peer'])

    def pack_args(self):
        return (self.format_str, (self.pad, self.peer))

    def unpack(str):
        return struct.unpack(PadPeerMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, pad, peer):
        return (pad, peer, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

__next_socket_id = 0

def get_socket_id():
    global __next_socket_id
    result = __next_socket_id
    __next_socket_id += 1
    return result

def unix_socketpair(type=socket.SOCK_DGRAM):
    "This function can be replaced by socket.socketpair in Python 2.4."
    sock0 = socket.socket(socket.AF_UNIX, type)
    name0 = struct.pack('b107s', 0, '/PipelineMC/%d/%d' %
                        (os.getpid(), get_socket_id()))
    sock0.bind(name0)
    sock1 = socket.socket(socket.AF_UNIX, type)
    name1 = struct.pack('b107s', 0, '/PipelineMC/%d/%d' % \
                        (os.getpid(), get_socket_id()))
    sock1.bind(name1)
    sock0.connect(name1)
    sock1.connect(name0)
    return (sock0, sock1)

class MsgFIFO (object):

    def __init__(self):
        self.cond = threading.Condition()
        self.fifo = []
        return

    def push(self, msg):
        self.cond.acquire()
        self.fifo.append(msg)
        self.cond.notify()
        self.cond.release()
        return

    def pop(self, timeout=None):
        self.cond.acquire()
        if len(self.fifo) == 0:
            self.cond.wait(timeout)
        if len(self.fifo) > 0:
            result = self.fifo.pop(0)
        else:
            result = None
        self.cond.release()
        return result

class SocketMonitor (threading.Thread):

    def __init__(self, name, ctl, udp=None, unix=None, pop_timeout=None):
        super(SocketMonitor, self).__init__(name=name)
        self.fifos = {}
        self.pop_timeout = pop_timeout
        self.lock = threading.Lock() # guards self.fifos
        self.udp = udp
        if self.udp is not None:
            self.udp.setblocking(0)
        self.unix = unix
        if self.unix is not None:
            self.unix.setblocking(0)
        self.ctl = ctl
        self.ctl.setblocking(0)
        return

    def push_msg(self, addr, msg):
        self.lock.acquire()
        try:
            fifo = self.fifos[addr]
        except KeyError:
            fifo = MsgFIFO()
            self.fifos[addr] = fifo
        self.lock.release()
        return fifo.push(msg)

    def pop_msg(self, addr, timeout=None):
        self.lock.acquire()
        try:
            fifo = self.fifos[addr]
        except KeyError:
            fifo = None
        self.lock.release()
        if fifo is not None:
            if timeout is None:
                timeout = self.pop_timeout
            result = fifo.pop(timeout)
        else:
            result = None
        return result

    def run(self):
        ctl = self.ctl.fileno()
        udp = None
        unix = None
        poll = select.poll()
        poll.register(ctl, select.POLLIN | select.POLLHUP)
        if self.udp is not None:
            udp = self.udp.fileno()
            poll.register(udp, select.POLLIN)
        if self.unix is not None:
            unix = self.unix.fileno()
            poll.register(unix, select.POLLIN)
        finished = False
        while not finished:
            try:
                events = poll.poll()
            except select.error, e:
                events = []
            for (fd, event) in events:
                if (event & select.POLLIN) != 0:
                    sock = None
                    if fd == ctl:
                        sock = self.ctl
                    elif fd == unix:
                        sock = self.unix
                    elif fd == udp:
                        sock = self.udp
                    assert sock is not None
                    (msg, addr) = sock.recvfrom(65536)
                    if fd == ctl:
                        finished = True
                    else:
                        self.push_msg(addr, msg)
                elif (event & (select.POLLHUP | select.POLLERR)) != 0:
                    finished = True
        return

class PipelineInstance:

    def __init__(self, family, addr=None, name=None):
        self.family = family
        self.addr = addr
        self.name = name
        self.__msgid = 0
        self.lock = threading.RLock()
        self.num_aliases = 0
        return

    def __next_msgid(self):
        result = self.__msgid
        self.__msgid = (self.__msgid + 1) % 65536
        return result

    msgid = property(lambda (self):self.__next_msgid())

    def acquire(self):
        return self.lock.acquire()

    def release(self):
        return self.lock.release()

class PipelineClient (object):

    input_stage = "INPUT_STAGE"
    timeout = 2.5

    def __init__(self, sockdir='/tmp', with_udp=True, with_unix=False):
        self.pipelines = Dict()
        self.sockdir = sockdir
        if with_unix:
            self.usock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
            # To get a reasonable name for our socket, I think that
            # the use of mktemp followed by bind would be sufficient;
            # but mktemp is deprecated, so we use mkstemp instead.
            # Unfortunately, mkstemp also produces an open file, which
            # is useless and must be closed and unlinked immediately
            # after opening. So there is a race condition here, but I
            # don't think that it's harmful.
            (tempfd, usockname) = mkstemp(dir=self.sockdir)
            os.close(tempfd)
            os.unlink(usockname)
            # the following raises socket.error if self.mysockname is in use
            self.usock.bind(usockname)
        else:
            self.usock = None
        if with_udp:
            self.isock =  socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
        else:
            self.isock = None
        (rctl, self.ctl) = unix_socketpair()
        self.ctl.setblocking(0)
        self.monitor = SocketMonitor(name='monitor', ctl = rctl,
                                     udp=self.isock,
                                     unix = self.usock,
                                     pop_timeout=self.timeout)
        self.monitor.setDaemon(True)
        self.monitor.start()
        return

    def __del__(self):
        self.ctl.send('X')
        self.monitor.join()
        if self.usock is not None:
            os.unlink(self.usock.getsockname())
        return

    def probe_addr(self, alias):
        probe = self.mc_transaction(alias=alias,
                                    msg=NameMsg(intent=MsgIntent.query))
        if probe is not None:
            result = probe.name
        else:
            result = None
        return result

    def get_instance(self, name, host, port):
        result = None
        family = None
        if host is None and port is None:
            family = socket.AF_UNIX
            addr = os.path.join(self.sockdir, name, 'ctlsock')
        elif host is not None and port is not None:
            family = socket.AF_INET
            addr = (socket.gethostbyname(host), port)
        if family is not None:
            for p in self.pipelines.values():
                if p.family == family and p.addr == addr:
                    if name is None or p.name == name:
                        result = p
            if result is None:
                result = PipelineInstance(family=family, addr=addr, name=name)
            result.acquire()
        return result

    def has_pipeline(self, alias):
        return alias in self.pipelines

    def add_pipeline(self, alias=None, name=None, host=None,
                     port=None, allow_new=True):
        result = False
        if alias is not None:
            # first create or find current PipelineInstance
            self.pipelines.acquire()
            pline = self.get_instance(name=name, host=host, port=port)
            if pline is not None:
                do_probe = False
                # pline was created, add it to self.pipelines
                # atomically to prevent a race condition with
                # another thread trying the same
                if alias not in self.pipelines:
                    self.pipelines[alias] = pline
                    self.pipelines.release()
                    pline.num_aliases += 1
                    # send a NameMsg probe
                    probe = self.probe_addr(alias)
                    if probe is not None:
                        if name is not None:
                            if name == probe:
                                result = True
                            else:
                                print 'Pipeline instance exists at %s, ' \
                                    'but the reported name is %s, ' \
                                    'not %s' % \
                                    (str(pline.addr), probe, name)
                        else:
                            if pline.name is None:
                                pline.name = probe
                            result = True
                    elif allow_new:
                        # probe failed, spawn new pipeline if allowed
                        args = ['wcbe']
                        if name is not None:
                            args.extend(['--name', name])
                        if pline.family == socket.AF_UNIX:
                            args.extend(['--sockdir', self.sockdir])
                            os.spawnvp(os.P_NOWAIT, 'wcbe', args)
                            do_probe = True
                        else: # pline.family == socket.AF_INET
                            args.extend(['--ctladdr',
                                         '%s:%d' % (pline.addr[0],
                                                    pline.addr[1])])
                            rc = os.system('ssh -x %s %s&' %
                                           (pline.addr[0], ' '.join(args)))
                            if rc == 0:
                                do_probe = True
                else:
                    # Alias already exists, verify that method
                    # arguments agree with pipeline instance
                    # parameters.
                    instance = self.pipelines[alias]
                    self.pipelines.release()
                    # Acquire the lock on the "current" pipeline
                    # instance to block in case another thread is in
                    # the process of spawning it.
                    instance.acquire()
                    try:
                        if instance.family == socket.AF_UNIX:
                            result = (host is None and port is None and
                                      os.path.join(self.sockdir, name,
                                                   'ctlsock') == instance.addr)
                        else: # pline.family = socket.AF_INET
                            result = ((socket.gethostbyname(host), port) ==
                                      instance.addr and
                                      (name is None or instance.name == name))
                    except:
                        pass
                    instance.release()

                if do_probe:
                    # send name probes to determine that pipeline
                    # is alive
                    t0 = time()
                    while time() - t0 < 15 and not result:
                        sleep(0.5)
                        probe = self.probe_addr(alias)
                        if probe is not None:
                            pline.name = probe
                            result = True

                # atomically remove PipelineInstance from
                # self.pipelines in case pipeline didn't start
                instance = None
                self.pipelines.acquire()
                if not result and alias in self.pipelines:
                    instance = self.pipelines[alias]
                    del self.pipelines[alias]
                self.pipelines.release()
                if instance is not None:
                    instance.acquire()
                    instance.num_aliases -= 1
                    instance.release()

                pline.release()

            else:
                self.pipelines.release()
        return result

    def remove_pipeline(self, alias, allow_quit=True):
        try:
            pline = self.pipelines[alias]
        except KeyError:
            pline = None
        if pline is not None:
            pline.acquire()
            pline.num_aliases -= 1
            if pline.num_aliases == 0 and allow_quit:
                self.send_msg(alias, QuitMsg())
            self.pipelines.acquire()
            if alias in self.pipelines:
                del self.pipelines[alias]
            self.pipelines.release()
            pline.release()
        return

    def send_msg(self, alias, msg):
        result = -1
        try:
            pline = self.pipelines[alias]
            if pline.family == socket.AF_UNIX:
                sock = self.usock
            else:
                sock = self.isock
            if pline.addr is not None:
                msg.id = pline.msgid
                sock.sendto(msg.pack(), pline.addr)
                result = msg.id
        except socket.error, e:
            if e[0] != errno.ENOENT:
                raise socket.error(e)
        return result

    def recv_msg(self, alias):
        msg = None
        pline = self.pipelines[alias]
        if pline.addr is not None:
            msg = self.monitor.pop_msg(pline.addr)
            if msg is not None:
                msg = CbeMsg.unpack(msg)
        return msg

    def mc_transaction(self, alias, msg):
        result = None
        if msg.direction == MsgDirection.request:
            pline = self.pipelines[alias]
            pline.acquire()
            try:
                msg_id = self.send_msg(alias, msg)
                if msg_id != -1 and msg.intent == MsgIntent.query:
                    t0 = time()
                    result_id = msg_id - 1
                    while time() - t0 < self.timeout and result_id != msg_id:
                        result = self.recv_msg(alias)
                        if result is not None:
                            result_id = result.id
            finally:
                pline.release()
        return result
