#include <glib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <gst/gst.h>

#include <term_source.h>
#include <enums.h>
#include <wcbe.h>
#include <wcbe_lagset_src.h>

#define LAGSET_SRC_NAME "src"
#define INPUT_STAGE_SOCK_NAME "issink"

typedef struct pipeline {
    GstElement *pipeline;
    gchar *input_stage;
    gchar *name;
    gint sock;
    gboolean finished;
} pipeline_t;

static void parse_options(gint argc, gchar *argv[], gchar **input,
                          gchar ***pipeline_desc);
static void run_pipeline(pipeline_t *pipeline);
static gboolean bus_cb(GstBus *bus, GstMessage *msg, gpointer data);
static void forward_message_upstream(pipeline_t *pipeline,
                                     GstStructure *structure);
static void handle_stream_msg(pipeline_t *pipeline,
                              const GstStructure *structure);
static void element_list_query(pipeline_t *pipeline, GstStructure *structure);
static void property_list_query(pipeline_t *pipeline, GstStructure *structure);
static void property_query(pipeline_t *pipeline, GstStructure *structure);
static void property_command(pipeline_t *pipeline, GstStructure *structure);
static void pad_list_query(pipeline_t *pipeline, GstStructure *structure);
static void pad_caps_query(pipeline_t *pipeline, GstStructure *structure);
static void pad_peer_query(pipeline_t *pipeline, GstStructure *structure);

gint
main(gint argc, gchar *argv[])
{
    gchar *input = NULL;
    gchar **pipeline_desc = NULL;
    pipeline_t pipeline = {.pipeline = NULL, .sock = -1};

    if (!g_thread_supported()) {
        g_thread_init(NULL);
    }
    enums_register();

    parse_options(argc, argv, &input, &pipeline_desc);
    GPtrArray *full_pipeline = g_ptr_array_new();
    g_ptr_array_add(full_pipeline, g_strdup("wcbe_lagset_src"));
    g_ptr_array_add(full_pipeline, g_strdup("name="LAGSET_SRC_NAME));
    gchar *insock = g_strdup_printf("input=%s", input);
    g_ptr_array_add(full_pipeline, insock);
    g_ptr_array_add(full_pipeline, g_strdup("!"));
    gchar **component = pipeline_desc;
    while (component != NULL && *component != NULL) {
        g_ptr_array_add(full_pipeline, g_strdup(*component));
        ++component;
    }
    g_ptr_array_add(full_pipeline, NULL);
    g_free(pipeline_desc);

    GError *err = NULL;
    pipeline.pipeline =
        gst_parse_launchv((const gchar **)full_pipeline->pdata, &err);
    if (err != NULL) {
        g_error("Failed to create pipeline: %s", err->message);
        if (pipeline.pipeline != NULL) {
            gst_object_unref(GST_OBJECT(pipeline.pipeline));
            pipeline.pipeline = NULL;
        }
    }
    gchar **pdata = (gchar **)g_ptr_array_free(full_pipeline, FALSE);
    g_strfreev(pdata);

    if (pipeline.pipeline != NULL) {
        // TODO: not sure of the encoding the g_path... functions work
        // on; ignored for now, need to experiment
        gchar *basename = g_path_get_basename(input);
        gchar *confdir = g_path_get_dirname(input);
        gchar *confname = g_path_get_basename(confdir);
        gchar *bedir = g_path_get_dirname(confdir);
        pipeline.input_stage =
            g_build_filename(bedir, INPUT_STAGE_SOCK_NAME, NULL);
        pipeline.name = g_build_filename(confname, basename, NULL);
        gst_element_set_name(pipeline.pipeline, basename);
        g_free(bedir);
        g_free(confname);
        g_free(confdir);
        g_free(basename);
        pipeline.sock = socket(PF_UNIX, SOCK_DGRAM, 0);
        if (pipeline.sock != -1) {
            gint rc = fcntl(pipeline.sock, F_SETFL, O_NONBLOCK);
            if (rc != 0) {
                GST_ERROR("Failed to set socket to non-blocking mode: %s",
                          g_strerror(errno));
            }
            run_pipeline(&pipeline);
            close(pipeline.sock);
        }
        else {
            g_printerr("Failed to open message passing socket: %s",
                       g_strerror(errno));
        }
        gst_object_unref(GST_OBJECT(pipeline.pipeline));
        g_free(pipeline.input_stage);
        g_free(pipeline.name);
    }
    return 0;
}

static void
parse_options(gint argc, gchar *argv[], gchar **input, gchar ***pipeline_desc)
{
    const GOptionEntry entries[] = {
        {"input", 'i', 0, G_OPTION_ARG_STRING, input,
         "Lag set source socket name", "NAME"},
        {NULL}
    };

    GError *error = NULL;
    GOptionContext *context = g_option_context_new(
        "PIPELINE - WIDAR correlator backend lag set pipeline");
    g_option_context_set_summary(
        context,
        "Lag set processing pipeline stage for compute "
        "node of WIDAR correlator backend system.\n"
        "PIPELINE argument should be given in the syntax "
        "used by gst-launch.");
    g_option_context_set_description(
        context,
        "Please report feature requests, bugs, etc. to\n"
        "mpokorny@nrao.edu.");
    g_option_context_set_help_enabled(context, TRUE);
    g_option_context_set_ignore_unknown_options(context, TRUE);
    g_option_context_add_main_entries(context, entries, NULL);
    g_option_context_add_group(context, gst_init_get_option_group());
    gboolean ctxt_parse =
        g_option_context_parse(context, &argc, &argv, &error);

    if (ctxt_parse) {
        GPtrArray *pipeline = g_ptr_array_new();
        guint i;
        for (i = 1; i < argc; ++i) {
            g_ptr_array_add(pipeline, argv[i]);
        }
        g_ptr_array_add(pipeline, NULL);
        *pipeline_desc = (gchar **)g_ptr_array_free(pipeline, FALSE);
    }

    if (!ctxt_parse) {
        if (error != NULL) {
            g_printerr("%s\n", error->message);
            g_error_free(error);
        }
        exit(-1);
    }
    g_option_context_free(context);
}

static void
run_pipeline(pipeline_t *pipeline)
{
    GstBus *bus = NULL;

    bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    if (bus == NULL) {
        g_printerr("Failed to get pipeline message bus");
        goto cleanup_and_return;
    }
    gst_bus_add_watch(bus, bus_cb, pipeline);

    GstStateChangeReturn ret =
        gst_element_set_state(pipeline->pipeline, GST_STATE_PLAYING);
    if (ret != GST_STATE_CHANGE_FAILURE) {
        GMainLoop *loop = g_main_loop_new(NULL, FALSE);

        // set up TermSource to catch termination signals
        pipeline->finished = FALSE;
        gboolean term_handler(gint *signum) {
            pipeline->finished = TRUE;
            return FALSE;
        }
        TermSource *term_source = term_source_new();
        term_source_set_callback(term_source, term_handler);
        term_source_attach(term_source, NULL);

        // On interrupt, try an orderly pipeline shutdown, then quit
        // main loop.
        gboolean done(gpointer data) {
            if (g_atomic_int_get(&(pipeline->finished))) {
                ret = gst_element_set_state(pipeline->pipeline,
                                            GST_STATE_NULL);
                if (ret == GST_STATE_CHANGE_ASYNC) {
                    gst_element_get_state(pipeline->pipeline, NULL, NULL,
                                          10 * GST_SECOND);
                }
                GstMessage *msg = gst_message_new_application(
                    NULL,
                    gst_structure_new(
                        "pipeline-state",
                        "value", PIPELINE_STATE_ENUM_TYPE,
                        PIPELINE_STATE_STOPPED,
                        NULL));
                if (!bus_cb(bus, msg, pipeline)) {
                    g_debug("Failed to post 'pipeline-state:value=stopped' "
                            "message to message bus");
                }
                gst_message_unref(msg);
                g_main_loop_quit(loop);
            }
            return TRUE;
        };
        g_timeout_add(100, done, NULL);

        GstMessage *msg = gst_message_new_application(
            NULL,
            gst_structure_new(
                "pipeline-state",
                "value", PIPELINE_STATE_ENUM_TYPE, PIPELINE_STATE_RUNNING,
                NULL));

        if (!bus_cb(bus, msg, pipeline)) {
            g_debug("Failed to post 'pipeline-state:value=running' message "
                    "to message bus");
        }
        gst_message_unref(msg);

        g_main_loop_run(loop);

        g_main_loop_unref(loop);
    }
    else {
        g_warning("Failed to start pipeline");
        GstMessage *msg = gst_bus_poll(bus, GST_MESSAGE_ERROR, 0);
        if (msg != NULL) {
            GError *error = NULL;
            gst_message_parse_error(msg, &error, NULL);
            g_printerr("ERROR: %s\n", error->message);
            g_error_free(error);
            gst_message_unref(msg);
        }
    }

 cleanup_and_return:
    if (bus != NULL) {
        gst_object_unref(GST_OBJECT(bus));
    }
}

static gboolean
bus_cb(GstBus *bus, GstMessage *msg, gpointer data)
{
    pipeline_t *pipeline = data;

    switch (GST_MESSAGE_TYPE(msg)) {
      case GST_MESSAGE_ERROR: {
        gchar *debug = NULL;
        GError *err = NULL;
        gst_message_parse_error(msg, &err, &debug);
        g_printerr("ERROR: %s\n", err->message);
        if (debug != NULL) {
            g_printerr("Debug details: %s\n", debug);
            g_free(debug);
        }
        g_error_free(err);
        break;
      }
      case GST_MESSAGE_EOS:
        g_atomic_int_set(&(pipeline->finished), TRUE);
        break;
      case GST_MESSAGE_APPLICATION: {
        const GstStructure *structure = gst_message_get_structure(msg);
        if (GST_MESSAGE_SRC(msg) != NULL) {
            gchar *src_name = gst_object_get_name(GST_MESSAGE_SRC(msg));
            if (strcmp(LAGSET_SRC_NAME, src_name) == 0
                && gst_structure_has_name(structure, STREAM_MSG_NAME)) {
                handle_stream_msg(pipeline, structure);
            }
            else {
                gchar *src = g_strjoin(G_DIR_SEPARATOR_S, pipeline->name,
                                       src_name, NULL);
                GstStructure *new_structure = gst_structure_copy(structure);
                gst_structure_set(new_structure,
                                  "src", G_TYPE_STRING, src,
                                  NULL);
                forward_message_upstream(pipeline, new_structure);
                g_free(src);
            }
            g_free(src_name);
        }
        else {
            GstStructure *new_structure = gst_structure_copy(structure);
            gst_structure_set(new_structure,
                              "src", G_TYPE_STRING, pipeline->name,
                              NULL);
            forward_message_upstream(pipeline, new_structure);
        }
        break;
      }
      default:
        break;
    }
    return TRUE;
}

static void
forward_message_upstream(pipeline_t *pipeline, GstStructure *structure)
{
    gchar *str = gst_structure_to_string(structure);
    guint str_len = strlen(str) + 1;
    struct sockaddr_un addr = {.sun_family = AF_UNIX};
    gint rc = g_strlcpy(addr.sun_path, pipeline->input_stage,
                        sizeof(addr.sun_path));
    g_assert(0 < rc && rc < sizeof(addr.sun_path));
    struct iovec iovec[] =
        {{.iov_base = &str_len, .iov_len = sizeof(str_len)},
         {.iov_base = str, .iov_len = str_len}};
    struct msghdr msghdr = {.msg_name = &addr,
                            .msg_namelen = sizeof(addr),
                            .msg_control = NULL, .msg_controllen = 0,
                            .msg_iov = iovec, .msg_iovlen = 2};
    gssize bytes_written = sendmsg(pipeline->sock, &msghdr, 0);
    if (bytes_written != iovec[0].iov_len + iovec[1].iov_len) {
        g_warning("Failed to write application message (%s) to socket: %s",
                  str, g_strerror(errno));
    }
    g_free(str);
    gst_structure_free(structure);
}

static void
handle_stream_msg(pipeline_t *pipeline, const GstStructure *structure)
{
    if (gst_structure_has_field(structure, "msg")) {
        const gchar *msg = gst_structure_get_string(structure, "msg");
        GstStructure *msg_structure = gst_structure_from_string(msg, NULL);
        if (msg_structure != NULL) {
            const gchar *name = gst_structure_get_name(msg_structure);
            if (strcmp(name, ELEMENT_LIST_MSG_NAME) == 0) {
                element_list_query(pipeline, msg_structure);
            }
            else if (strcmp(name, PROPERTY_LIST_MSG_NAME) == 0) {
                property_list_query(pipeline, msg_structure);
            }
            else if (strcmp(name, GET_PROPERTY_MSG_NAME) == 0) {
                property_query(pipeline, msg_structure);
            }
            else if (strcmp(name, SET_PROPERTY_MSG_NAME) == 0) {
                property_command(pipeline, msg_structure);
            }
            else if (strcmp(name, PAD_LIST_MSG_NAME) == 0) {
                pad_list_query(pipeline, msg_structure);
            }
            else if (strcmp(name, PAD_CAPS_MSG_NAME) == 0) {
                pad_caps_query(pipeline, msg_structure);
            }
            else if (strcmp(name, PAD_PEER_MSG_NAME) == 0) {
                pad_peer_query(pipeline, msg_structure);
            }
            else {
                gst_structure_free(msg_structure);
            }
        }
    }
}

static void
element_list_query(pipeline_t *pipeline, GstStructure *structure)
{
    GString *names = g_string_new("");
    GstIterator *elements =
        gst_bin_iterate_recurse(GST_BIN(pipeline->pipeline));
    void element_name(gpointer data, gpointer user_data) {
        GstElement *element = GST_ELEMENT(data);
        gchar *name = gst_element_get_name(element);
        gst_object_unref(GST_OBJECT(element));
        if (names->len > 0) {
            g_string_append_printf(names, ",%s", name);
        }
        else {
            g_string_append(names, name);
        }
        g_free(name);
    }
    GstIteratorResult iter_result =
        gst_iterator_foreach(elements, element_name, NULL);
    gst_iterator_free(elements);
    g_assert(iter_result != GST_ITERATOR_OK);
    if (iter_result == GST_ITERATOR_DONE) {
        gst_structure_set(structure, "list", G_TYPE_STRING, names->str, NULL);
    }
    else {
        gst_structure_set(structure, "list", G_TYPE_STRING, "", NULL);
    }
    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    GstMessage *retmsg = gst_message_new_application(NULL, structure);
    g_assert(retmsg != NULL);
    if (!gst_bus_post(bus, retmsg)) {
        gst_message_unref(retmsg);
    }
    gst_object_unref(GST_OBJECT(bus));
    g_string_free(names, TRUE);
}

static void
property_list_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get list of element properties
    const gchar *element_name =
        gst_structure_get_string(structure, "element");
    GString *properties = g_string_new("");
    GstElement *element = gst_bin_get_by_name(GST_BIN(pipeline->pipeline),
                                              element_name);
    if (element != NULL) {
        guint num_properties;
        GParamSpec **params = g_object_class_list_properties(
            G_OBJECT_GET_CLASS(G_OBJECT(element)),
            &num_properties);
        guint i;
        for (i = 0; i < num_properties; ++i) {
            GParamSpec *param = params[i];
            const gchar *fmt;
            if (properties->len == 0) {
                fmt = "%s<%s>";
            }
            else {
                fmt = ",%s<%s>";
            }
            g_string_append_printf(properties, fmt, param->name,
                                   g_type_name(param->value_type));
        }
        g_free(params);
        gst_object_unref(GST_OBJECT(element));
    }
    gst_structure_set(structure, "list", G_TYPE_STRING, properties->str, NULL);
    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    GstMessage *retmsg = gst_message_new_application(NULL, structure);
    g_assert(retmsg != NULL);
    if (!gst_bus_post(bus, retmsg)) {
        gst_message_unref(retmsg);
    }
    gst_object_unref(GST_OBJECT(bus));
    g_string_free(properties, TRUE);
}

static void
property_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get property value
    const gchar *element_property_name =
        gst_structure_get_string(structure, "property");
    gchar **name_parts =
        g_strsplit(element_property_name, G_DIR_SEPARATOR_S, 0);
    if (g_strv_length(name_parts) == 2) {
        GstElement *element = gst_bin_get_by_name(GST_BIN(pipeline->pipeline),
                                                  name_parts[0]);
        if (element != NULL) {
            GParamSpec *param = g_object_class_find_property(
                G_OBJECT_GET_CLASS(G_OBJECT(element)), name_parts[1]);
            if (param != NULL) {
                switch (param->value_type) {
                  case G_TYPE_CHAR: {
                    gchar v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_CHAR, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_UCHAR: {
                    guchar v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_UCHAR, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_BOOLEAN: {
                    gboolean v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_BOOLEAN, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_INT: {
                    gint v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_INT, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_UINT: {
                    guint v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_UINT, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_LONG: {
                    glong v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_LONG, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_ULONG: {
                    gulong v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_ULONG, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_INT64: {
                    gint64 v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_INT64, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_UINT64: {
                    guint64 v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_UINT64, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_FLOAT: {
                    gfloat v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_FLOAT, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_DOUBLE: {
                    gdouble v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_DOUBLE, v,
                                      NULL);
                    break;
                  }
                  case G_TYPE_STRING: {
                    gchar *v;
                    g_object_get(G_OBJECT(element), name_parts[1], &v, NULL);
                    gst_structure_set(structure, "value", G_TYPE_STRING, v,
                                      NULL);
                    break;
                  }
                  default:
                    if (g_type_is_a(param->value_type, G_TYPE_ENUM)) {
                        gint v;
                        g_object_get(G_OBJECT(element), name_parts[1], &v,
                                     NULL);
                        gst_structure_set(structure, "value",
                                          param->value_type, v, NULL);
                    }
                    else if (g_type_is_a(param->value_type, G_TYPE_FLAGS)) {
                        guint v;
                        g_object_get(G_OBJECT(element), name_parts[1], &v,
                                     NULL);
                        gst_structure_set(structure, "value",
                                          param->value_type, v, NULL);
                    }
                    break;
                }
            }
            gst_object_unref(GST_OBJECT(element));
        }
    }
    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    GstMessage *retmsg = gst_message_new_application(NULL, structure);
    g_assert(retmsg != NULL);
    if (!gst_bus_post(bus, retmsg)) {
        gst_message_unref(retmsg);
    }
    gst_object_unref(GST_OBJECT(bus));
}

static void
property_command(pipeline_t *pipeline, GstStructure *structure)
{
    // set element property value
    const gchar *element_property_name =
        gst_structure_get_string(structure, "property");
    gchar **name_parts =
        g_strsplit(element_property_name, G_DIR_SEPARATOR_S, 0);
    if (g_strv_length(name_parts) == 2) {
        GstElement *element = gst_bin_get_by_name(GST_BIN(pipeline->pipeline),
                                                  name_parts[0]);
        if (element != NULL) {
            GParamSpec *param = g_object_class_find_property(
                G_OBJECT_GET_CLASS(G_OBJECT(element)), name_parts[1]);
            const GValue *value = gst_structure_get_value(structure, "value");
            if (param != NULL && value != NULL) {
                g_object_set_property(G_OBJECT(element), name_parts[1], value);
            }
            gst_object_unref(GST_OBJECT(element));
        }
    }
}

static void
pad_list_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get list of element pads
    const gchar *element_name =
        gst_structure_get_string(structure, "element");
    GString *pads = g_string_new("");
    GstElement *element = gst_bin_get_by_name(GST_BIN(pipeline->pipeline),
                                              element_name);
    if (element != NULL) {
        GstIterator *pads_iter = gst_element_iterate_pads(element);
        void pad_name(gpointer data, gpointer user_data) {
            GstPad *pad = GST_PAD(data);
            gchar *name = gst_pad_get_name(pad);
            gst_object_unref(GST_OBJECT(pad));
            if (pads->len == 0) {
                g_string_append(pads, name);
            }
            else {
                g_string_append_printf(pads, ",%s", name);
            }
            g_free(name);
        }
        GstIteratorResult iter_result =
            gst_iterator_foreach(pads_iter, pad_name, NULL);
        gst_iterator_free(pads_iter);
        if (iter_result == GST_ITERATOR_ERROR) {
            g_string_truncate(pads, 0);
        }
        gst_object_unref(GST_OBJECT(element));
    }
    gst_structure_set(structure, "list", G_TYPE_STRING, pads->str, NULL);
    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    GstMessage *retmsg = gst_message_new_application(NULL, structure);
    g_assert(retmsg != NULL);
    if (!gst_bus_post(bus, retmsg)) {
        gst_message_unref(retmsg);
    }
    gst_object_unref(GST_OBJECT(bus));
    g_string_free(pads, TRUE);
}

static void
pad_caps_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get pad caps
    gchar *str = NULL;
    const gchar *pad_name = gst_structure_get_string(structure, "pad");
    gchar **name_parts = g_strsplit(pad_name, G_DIR_SEPARATOR_S, 0);
    if (g_strv_length(name_parts) == 2) {
        GstElement *element =
            gst_bin_get_by_name(GST_BIN(pipeline->pipeline), name_parts[0]);
        if (element != NULL) {
            GstPad *pad = gst_element_get_static_pad(element, name_parts[1]);
            if (pad != NULL) {
                GstCaps *caps = gst_pad_get_negotiated_caps(pad);
                if (caps != NULL) {
                    str = gst_caps_to_string(caps);
                    gst_caps_unref(caps);
                }
                gst_object_unref(GST_OBJECT(pad));
            }
            gst_object_unref(GST_OBJECT(element));
        }
    }
    if (str == NULL) {
        str = g_strdup("");
    }
    gst_structure_set(structure, "caps", G_TYPE_STRING, str, NULL);
    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    GstMessage *retmsg = gst_message_new_application(NULL, structure);
    g_assert(retmsg != NULL);
    if (!gst_bus_post(bus, retmsg)) {
        gst_message_unref(retmsg);
    }
    gst_object_unref(GST_OBJECT(bus));
    g_free(str);
}

static void
pad_peer_query(pipeline_t *pipeline, GstStructure *structure)
{
    // get pad peer
    gchar *str = NULL;
    const gchar *pad_name = gst_structure_get_string(structure, "pad");
    gchar **name_parts = g_strsplit(pad_name, G_DIR_SEPARATOR_S, 0);
    if (g_strv_length(name_parts) == 2) {
        GstElement *element =
            gst_bin_get_by_name(GST_BIN(pipeline->pipeline), name_parts[0]);
        if (element != NULL) {
            GstPad *pad = gst_element_get_static_pad(element, name_parts[1]);
            if (pad != NULL && gst_pad_is_linked(pad)) {
                GstPad *peer = gst_pad_get_peer(pad);
                g_assert(peer != NULL);
                GstObject *parent = gst_pad_get_parent(peer);
                g_assert(parent != NULL);
                str = g_strdup_printf("%s/%s", GST_OBJECT_NAME(parent),
                                      GST_OBJECT_NAME(peer));
                gst_object_unref(parent);
                gst_object_unref(GST_OBJECT(peer));
                gst_object_unref(GST_OBJECT(pad));
            }
            gst_object_unref(GST_OBJECT(element));
        }
    }
    if (str == NULL) {
        str = g_strdup("");
    }
    gst_structure_set(structure, "peer", G_TYPE_STRING, str, NULL);
    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    GstMessage *retmsg = gst_message_new_application(NULL, structure);
    g_assert(retmsg != NULL);
    if (!gst_bus_post(bus, retmsg)) {
        gst_message_unref(retmsg);
    }
    gst_object_unref(GST_OBJECT(bus));
    g_free(str);
}
