#include <term_source.h>

#include <signal.h>
#include <time.h>

static gint last_signal = 0;
static struct timespec sig_timestamp = {.tv_sec = 0, .tv_nsec = 0};
static gint ref_count = 0;

static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

static struct sigaction orig_sigterm;
static struct sigaction orig_sighup;
static struct sigaction orig_sigint;
static struct sigaction term_sigaction;

static void
term_handler(gint sig)
{
    last_signal = sig;
    clock_gettime(CLOCK_MONOTONIC, &sig_timestamp);
}

static gboolean
prepare(GSource *source, gint *timeout)
{
    gboolean result;
    TermSource *tsource = (TermSource *)source;
    if (sig_timestamp.tv_sec != tsource->sig_timestamp.tv_sec
        || sig_timestamp.tv_nsec != tsource->sig_timestamp.tv_nsec) {
        result = TRUE;
        *timeout = 0;
    }
    else {
        result = FALSE;
    }
    return result;
}

static gboolean
check(GSource *source)
{
    gint t;
    return prepare(source, &t);
}

static gboolean
dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
    gboolean result;
    TermSource *tsource = (TermSource *)source;
    tsource->sig_timestamp = sig_timestamp;
    if (callback != NULL) {
        gint sig = last_signal;
        result = callback(&sig);
    }
    else {
        result = TRUE;
    }
    return result;
}

static void
finalize(GSource *source)
{
    g_static_mutex_lock(&mutex);
    --ref_count;
    if (ref_count == 0) {
        sigaction(SIGTERM, &orig_sigterm, NULL);
        sigaction(SIGHUP, &orig_sighup, NULL);
        sigaction(SIGINT, &orig_sigint, NULL);
    }
    g_static_mutex_unlock(&mutex);
}

static GSourceFuncs term_funcs = {
    .prepare = prepare,
    .check = check,
    .dispatch = dispatch,
    .finalize = finalize
};

TermSource *
term_source_new(void)
{
    g_static_mutex_lock(&mutex);
    if (ref_count == 0) {
        term_sigaction.sa_handler = term_handler;
        sigaction(SIGTERM, &term_sigaction, &orig_sigterm);
        sigaction(SIGHUP, &term_sigaction, &orig_sighup);
        sigaction(SIGINT, &term_sigaction, &orig_sigint);
    }
    ++ref_count;
    g_static_mutex_unlock(&mutex);
    TermSource *result =
        (TermSource *)g_source_new(&term_funcs, sizeof(TermSource));
    result->sig_timestamp = sig_timestamp;
    return result;
}

void
term_source_set_callback(TermSource *source, TermSourceFunc func)
{
    g_source_set_callback((GSource *)source, (GSourceFunc)func, NULL, NULL);
}
