#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <wcbe_mc.h>
#include <wcbe_framedump.h>

#ifdef G_HAVE_ISO_VARARGS
#define g_info(...)    g_log (G_LOG_DOMAIN,       \
                              G_LOG_LEVEL_INFO,   \
                              __VA_ARGS__)
#elif defined(G_HAVE_GNUC_VARARGS)
#define g_info(format...)      g_log (G_LOG_DOMAIN,         \
                                      G_LOG_LEVEL_INFO,     \
                                      format)
#else   /* no varargs macros */
static void
g_info (const gchar *format,
        ...)
{
  va_list args;
  va_start (args, format);
  g_logv (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, format, args);
  va_end (args);
}
#endif  /* !__GNUC__ */

static void concat_string_list(GPtrArray *strings, gchar *result,
                               gboolean free_strings);
static request_cb_fn_t ls_message(
    pipeline_t *pipeline, const gchar *config_name,
    const gchar *ls_pipeline_name, const wcbe_msg_t *msg,
    GstStructure *structure, request_cb_fn_t cb);
static gboolean forward_message(
    pipeline_t *pipeline, const gchar *config_name,
    const gchar *ls_pipeline_name, GstStructure *structure);
static configuration_t *find_config(pipeline_t *pipeline, const gchar *name);
static void element_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                               gsize *msgsize, GstStructure *structure);
static void property_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                                gsize *msgsize, GstStructure *structure);
static void property_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize, GstStructure *structure);
static void pad_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize, GstStructure *structure);
static void pad_caps_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize, GstStructure *structure);
static void pad_peer_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize, GstStructure *structure);

const gchar wcbe_input_stage_pipeline_name[] = "INPUT_STAGE";

request_cb_fn_t
wcbe_mc_config_create_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                              gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_CREATE_MSG, NULL);

    wcbe_config_create_msg_t *create_msg = (wcbe_config_create_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, create_msg->name);
    if (config == NULL) {
        gboolean free_config = FALSE;
        config = g_new(configuration_t, 1);
        config->name = g_strdup(create_msg->name);
        config->desc = g_strdup(create_msg->desc);
        config->ls_pipelines = NULL;
        config->dir = g_build_filename(pipeline->sockdir, config->name, NULL);
        config->destroy = FALSE;
        config->pipeline = pipeline;
        gint rc = mkdir(config->dir, S_IRWXU);
        if (rc == 0) {
            if (pipeline->is_pipeline != NULL
                && pipeline->flow == WCBE_FLOW_ON) {
                free_config = !wcbe_create_config(pipeline, config);
            }
            else {
                wcbe_insert_pending_conf(pipeline, PENDING_CREATE, config);
            }
        }
        else {
            g_info("Failed to create configuration directory '%s': %s",
                   config->dir, g_strerror(errno));
            free_config = TRUE;
        }
        if (free_config) {
            wcbe_cleanup_config(config);
        }
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_config_activate_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                                gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_ACTIVATE_MSG, NULL);

    wcbe_config_activate_msg_t *activate_msg =
        (wcbe_config_activate_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, activate_msg->name);
    if (config != NULL) {
        if (pipeline->is_pipeline != NULL && pipeline->flow == WCBE_FLOW_ON) {
            if (!config->destroy
                && !wcbe_activate_config(pipeline, config)) {
                g_timeout_add(DELAYED_ACTIVATE_INTERVAL,
                              wcbe_delayed_activate_config, config);
            }
        }
        else {
            wcbe_insert_pending_conf(pipeline, PENDING_ACTIVATE, config);
        }
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_config_destroy_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                               gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_DESTROY_MSG, NULL);

    wcbe_config_destroy_msg_t *destroy_msg = (wcbe_config_destroy_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, destroy_msg->name);
    if (config != NULL) {
        if (pipeline->is_pipeline != NULL && pipeline->flow == WCBE_FLOW_ON) {
            wcbe_destroy_config(pipeline, config);
        }
        else {
            wcbe_insert_pending_conf(pipeline, PENDING_DESTROY, config);
        }
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_config_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                          gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_LIST_MSG, NULL);

    wcbe_config_list_msg_t **config_list_msg = (wcbe_config_list_msg_t **)msg;
    GPtrArray *configs = g_ptr_array_new();
    (*config_list_msg)->list_len = 1;
    // configs in active pipeline
    void config_name(gpointer key, gpointer value, gpointer user_data) {
        const gchar *name = key;
        const configuration_t *cfg = value;
        const gchar *fmt;
        if (cfg->destroy) {
            fmt = "[%s]";
        }
        else if (cfg == pipeline->config) {
            fmt = "<%s>";
        }
        else {
            fmt = "(%s)";
        }
        gchar *tagged_name = g_strdup_printf(fmt, name);
        g_ptr_array_add(configs, tagged_name);
        (*config_list_msg)->list_len += strlen(tagged_name) + 1;
    }
    g_hash_table_foreach(pipeline->configs, config_name, NULL);
    // pending configs
    void pending_config_name(gpointer data, gpointer user_data) {
        pending_config_t *pconf = data;
        const gchar *fmt;
        switch (pconf->action) {
          case PENDING_CREATE:
            fmt = "*%s*";
            break;
          case PENDING_ACTIVATE:
            fmt = "#%s#";
            break;
          case PENDING_DESTROY:
            fmt = "@%s@";
            break;
          default:
            g_assert_not_reached();
            break;
        }
        gchar *tagged_name = g_strdup_printf(fmt, pconf->config->name);
        g_ptr_array_add(configs, tagged_name);
        (*config_list_msg)->list_len += strlen(tagged_name) + 1;
    }
    g_slist_foreach(pipeline->pending_configs, pending_config_name, NULL);

    *msgsize = sizeof(wcbe_config_list_msg_t) + (*config_list_msg)->list_len;
    *config_list_msg = g_realloc(*config_list_msg, *msgsize);
    concat_string_list(configs, (*config_list_msg)->list, TRUE);
    g_ptr_array_free(configs, TRUE);
    return NULL;
}

request_cb_fn_t
wcbe_mc_src_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_SRC_MSG, NULL);

    wcbe_src_msg_t *src_msg = (wcbe_src_msg_t *)*msg;
    src_msg->name[sizeof(src_msg->name) - 1] = '\0';

    if (pipeline->src == NULL
        || strcmp(src_msg->name, pipeline->src) != 0) {
        wcbe_flow_t flow = pipeline->flow;
        wcbe_destroy_pipeline(pipeline);
        pipeline->src = g_strdup(src_msg->name);
        if (pipeline->src[0] != '\0') {
            wcbe_create_pipeline(pipeline);
            if (pipeline->is_pipeline != NULL && flow == WCBE_FLOW_ON) {
                wcbe_start_pipeline(pipeline, TRUE);
            }
        }
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_src_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_SRC_MSG, NULL);

    wcbe_src_msg_t *src_msg = (wcbe_src_msg_t *)*msg;
    const gchar *name = (pipeline->src ? pipeline->src : "");
    g_strlcpy(src_msg->name, name, sizeof(src_msg->name));
    return NULL;
}

request_cb_fn_t
wcbe_mc_live_ts_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_LIVE_TS_MSG, NULL);

    wcbe_live_ts_msg_t *live_ts_msg = (wcbe_live_ts_msg_t *)*msg;
    if (live_ts_msg->state != pipeline->live_ts) {
        if (pipeline->is_pipeline != NULL) {
            GstElement *src =
                gst_bin_get_by_name(GST_BIN(pipeline->is_pipeline), "src");
            g_assert(src != NULL);
            g_object_set(G_OBJECT(src), "live-timestamps", live_ts_msg->state,
                         NULL);
            gst_object_unref(GST_OBJECT(src));
        }
        pipeline->live_ts = live_ts_msg->state;
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_live_ts_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_LIVE_TS_MSG, NULL);

    wcbe_live_ts_msg_t *live_ts_msg = (wcbe_live_ts_msg_t *)*msg;
    live_ts_msg->state = pipeline->live_ts;
    return NULL;
}

request_cb_fn_t
wcbe_mc_framedump_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                          gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_FRAMEDUMP_MSG, NULL);

    wcbe_framedump_msg_t *framedump_msg = (wcbe_framedump_msg_t *)*msg;
    if (framedump_msg->format != pipeline->framedump) {
        if (pipeline->is_pipeline != NULL) {
            GstElement *framedump = gst_bin_get_by_name(
                GST_BIN(pipeline->is_pipeline), "framedump");
            g_assert(framedump != NULL);
            if (framedump_msg->format != WCBE_DUMP_NONE) {
                WcbeFramedumpFormat fdformat =
                    ((framedump_msg->format == WCBE_DUMP_BINARY)
                     ? WCBE_FRAMEDUMP_BINARY
                     : WCBE_FRAMEDUMP_ASCII);
                g_object_set(G_OBJECT(framedump), "do-output", FALSE, NULL);
                g_object_set(G_OBJECT(framedump), "format", fdformat, NULL);
                g_object_set(G_OBJECT(framedump), "do-output", TRUE, NULL);
            }
            else {
                g_object_set(G_OBJECT(framedump), "do-output", FALSE, NULL);
            }
            gst_object_unref(GST_OBJECT(framedump));
        }
        pipeline->framedump = framedump_msg->format;
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_framedump_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_FRAMEDUMP_MSG, NULL);

    wcbe_framedump_msg_t *framedump_msg = (wcbe_framedump_msg_t *)*msg;
    framedump_msg->format = pipeline->framedump;
    return NULL;
}

request_cb_fn_t
wcbe_mc_flow_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_FLOW_MSG, NULL);

    wcbe_flow_msg_t *flow_msg = (wcbe_flow_msg_t *)*msg;
    if (flow_msg->state != pipeline->flow) {
        if (flow_msg->state == WCBE_FLOW_ON) {
            if (pipeline->is_pipeline != NULL) {
                wcbe_start_pipeline(pipeline, TRUE);
            }
        }
        else { // flow_msg->state == WCBE_FLOW_OFF
            if (pipeline->is_pipeline != NULL) {
                wcbe_stop_pipeline(pipeline, TRUE);
            }
        }
        // start/stop_pipeline update pipeline->flow automatically
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_flow_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_FLOW_MSG, NULL);

    wcbe_flow_msg_t *flow_msg = (wcbe_flow_msg_t *)*msg;
    flow_msg->state = pipeline->flow;
    return NULL;
}

request_cb_fn_t
wcbe_mc_rate_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_RATE_MSG, NULL);

    wcbe_rate_msg_t *rate_msg = (wcbe_rate_msg_t *)*msg;
    if (rate_msg->fps != pipeline->fps) {
        if (pipeline->is_pipeline != NULL) {
            GstElement *src =
                gst_bin_get_by_name(GST_BIN(pipeline->is_pipeline), "src");
            g_assert(src != NULL);
            g_object_set(G_OBJECT(src), "fps", rate_msg->fps, NULL);
            gst_object_unref(GST_OBJECT(src));
        }
        pipeline->fps = rate_msg->fps;
    }
    return NULL;
}

request_cb_fn_t
wcbe_mc_rate_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_RATE_MSG, NULL);

    wcbe_rate_msg_t *rate_msg = (wcbe_rate_msg_t *)*msg;
    rate_msg->fps = pipeline->fps;
    return NULL;
}

request_cb_fn_t
wcbe_mc_name_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_NAME_MSG, NULL);

    wcbe_name_msg_t *name_msg = (wcbe_name_msg_t *)*msg;
    g_assert(pipeline->name != NULL);
    g_strlcpy(name_msg->value, pipeline->name, sizeof(name_msg->value));
    return NULL;
}

request_cb_fn_t
wcbe_mc_quit_command(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_QUIT_MSG, NULL);
    g_main_loop_quit(pipeline->loop);
    return NULL;
}

request_cb_fn_t
wcbe_mc_pipeline_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                            gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PIPELINE_LIST_MSG, NULL);

    wcbe_pipeline_list_msg_t **p_msg = (wcbe_pipeline_list_msg_t **)msg;
    (*p_msg)->list_len = 1;
    GPtrArray *pipelines = g_ptr_array_new();
    configuration_t *config =
        g_hash_table_lookup(pipeline->configs, (*p_msg)->configuration);
    if (config != NULL) {
        if (config == pipeline->config) {
            g_ptr_array_add(pipelines,
                            (gchar *)wcbe_input_stage_pipeline_name);
            (*p_msg)->list_len += strlen(wcbe_input_stage_pipeline_name) + 1;
        }
        void pipeline_name(gpointer data, gpointer user_data) {
            ls_pipeline_t *p = data;
            gchar *pname = strchr(p->name, G_DIR_SEPARATOR);
            g_assert(pname != NULL);
            pname += 1;
            g_ptr_array_add(pipelines, pname);
            (*p_msg)->list_len += strlen(pname) + 1;
        }
        g_slist_foreach(config->ls_pipelines, pipeline_name, NULL);
    }
    *msgsize = sizeof(wcbe_pipeline_list_msg_t) + (*p_msg)->list_len;
    *p_msg = g_realloc(*p_msg, *msgsize);
    concat_string_list(pipelines, (*p_msg)->list, FALSE);
    g_ptr_array_free(pipelines, TRUE);
    return NULL;
}

request_cb_fn_t
wcbe_mc_element_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                           gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_ELEMENT_LIST_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_element_list_msg_t **e_msg = (wcbe_element_list_msg_t **)msg;
    (*e_msg)->list_len = 0;
    gchar **name_parts = g_strsplit((*e_msg)->pipeline, G_DIR_SEPARATOR_S, 3);
    if (g_strv_length(name_parts) == 2) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        if (strcmp(config_name, pipeline->config->name) == 0
            && strcmp(pipeline_name, wcbe_input_stage_pipeline_name) == 0) {
            // elements in input stage pipeline
            if (pipeline->is_pipeline != NULL) {
                (*e_msg)->list_len = 1;
                GPtrArray *elements = g_ptr_array_new();
                GstIterator *children =
                    gst_bin_iterate_recurse(GST_BIN(pipeline->is_pipeline));
                void element_name(gpointer data, gpointer user_data) {
                    GstElement *element = GST_ELEMENT(data);
                    gchar *name = gst_element_get_name(element);
                    gst_object_unref(GST_OBJECT(element));
                    g_ptr_array_add(elements, name);
                    (*e_msg)->list_len += strlen(name) + 1;
                }
                GstIteratorResult iter_result =
                    gst_iterator_foreach(children, element_name, NULL);
                gst_iterator_free(children);
                if (iter_result == GST_ITERATOR_ERROR) {
                    guint i;
                    for (i = 0; i < elements->len; ++i) {
                        gchar *name = g_ptr_array_index(elements, i);
                        g_free(name);
                        (*e_msg)->list_len = 1;
                    }
                    elements->len = 0;
                }
                *msgsize =
                    sizeof(wcbe_element_list_msg_t) + (*e_msg)->list_len;
                *e_msg = g_realloc(*e_msg, *msgsize);
                concat_string_list(elements, (*e_msg)->list, TRUE);
                g_ptr_array_free(elements, TRUE);
            }
        }

        else {
            // elements (perhaps) in a lag processing stage pipeline
            GstStructure *structure =
                gst_structure_empty_new(ELEMENT_LIST_MSG_NAME);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, element_list_query);
        }
    }
    g_strfreev(name_parts);
    return result;
}

request_cb_fn_t
wcbe_mc_property_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                            gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PROPERTY_LIST_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_property_list_msg_t **p_msg = (wcbe_property_list_msg_t **)msg;
    (*p_msg)->list_len = 0;
    gchar **name_parts = g_strsplit((*p_msg)->element, G_DIR_SEPARATOR_S, 4);
    if (g_strv_length(name_parts) == 3) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        if (strcmp(config_name, pipeline->config->name) == 0
            && strcmp(pipeline_name, wcbe_input_stage_pipeline_name) == 0) {
            // input stage element
            if (pipeline->is_pipeline != NULL) {
                GPtrArray *properties = g_ptr_array_new();
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->is_pipeline), element_name);
                if (element != NULL) {
                    (*p_msg)->list_len = 1;
                    guint num_properties;
                    GParamSpec **params = g_object_class_list_properties(
                        G_OBJECT_GET_CLASS(G_OBJECT(element)),
                        &num_properties);
                    guint i;
                    for (i = 0; i < num_properties; ++i) {
                        GParamSpec *param = params[i];
                        gchar *name = g_strdup_printf(
                            "%s<%s>", param->name,
                            g_type_name(param->value_type));
                        g_ptr_array_add(properties, name);
                        (*p_msg)->list_len += strlen(name) + 1;
                    }
                    g_free(params);
                    gst_object_unref(GST_OBJECT(element));
                }
                *msgsize =
                    sizeof(wcbe_property_list_msg_t) + (*p_msg)->list_len;
                *p_msg = g_realloc(*p_msg, *msgsize);
                concat_string_list(properties, (*p_msg)->list, TRUE);
                g_ptr_array_free(properties, TRUE);
            }
        }
        else {
            // element (perhaps) in a lag processing stage pipeline
            GstStructure *structure =
                gst_structure_new(PROPERTY_LIST_MSG_NAME,
                                  "element", G_TYPE_STRING, element_name,
                                  NULL);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, property_list_query);
        }
    }
    g_strfreev(name_parts);

    return result;
}

request_cb_fn_t
wcbe_mc_property_command(pipeline_t *pipeline, wcbe_msg_t **msg,
                         gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PROPERTY_MSG, NULL);

    wcbe_property_msg_t *p_msg = (wcbe_property_msg_t *)*msg;
    gchar **name_parts = g_strsplit(p_msg->property, G_DIR_SEPARATOR_S, 5);
    if (g_strv_length(name_parts) == 4) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        const gchar *property_name = name_parts[3];
        if (strcmp(config_name, pipeline->config->name) == 0
            && strcmp(pipeline_name, wcbe_input_stage_pipeline_name) == 0) {
            // input stage element property
            if (pipeline->is_pipeline != NULL) {
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->is_pipeline), element_name);
                if (element != NULL) {
                    GParamSpec *param = g_object_class_find_property(
                        G_OBJECT_GET_CLASS(G_OBJECT(element)), property_name);
                    GType valtype = g_type_from_name(p_msg->type);
                    if (param != NULL && valtype != 0) {
                        switch (valtype) {
                          case G_TYPE_CHAR:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gint8 *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_UCHAR:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((guchar *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_BOOLEAN:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gint *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_INT:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gint *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_UINT:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((guint *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_LONG:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((glong *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_ULONG:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gulong *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_INT64:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gint64 *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_UINT64:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((guint64 *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_FLOAT:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gfloat *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_DOUBLE:
                            g_object_set(G_OBJECT(element), property_name,
                                         *((gdouble *)(p_msg->value)), NULL);
                            break;
                          case G_TYPE_STRING:
                            g_object_set(G_OBJECT(element), property_name,
                                         p_msg->value, NULL);
                            break;
                          default:
                            if (g_type_is_a(valtype, G_TYPE_ENUM)) {
                                g_object_set(G_OBJECT(element), property_name,
                                             *((gint *)(p_msg->value)), NULL);
                            }
                            else if (g_type_is_a(valtype, G_TYPE_FLAGS)) {
                                g_object_set(G_OBJECT(element), property_name,
                                             *((guint *)(p_msg->value)), NULL);
                            }
                            break;
                        }
                    }
                    gst_object_unref(GST_OBJECT(element));
                }
            }
        }
        else {
            // property of element (perhaps) in lag processing stage
            gchar *propname = g_strjoin(
                G_DIR_SEPARATOR_S, element_name, property_name, NULL);
            GstStructure *structure =
                gst_structure_new(SET_PROPERTY_MSG_NAME,
                                  "property", G_TYPE_STRING, propname,
                                  NULL);
            g_assert(structure != NULL);
            GType valtype = g_type_from_name(p_msg->type);
            switch (valtype) {
              case G_TYPE_CHAR:
                gst_structure_set(structure, "value", G_TYPE_CHAR,
                                  *(gint8 *)(p_msg->value), NULL);
                break;
              case G_TYPE_UCHAR:
                gst_structure_set(structure, "value", G_TYPE_UCHAR,
                                  *(guchar *)(p_msg->value), NULL);
                break;
              case G_TYPE_BOOLEAN:
                gst_structure_set(structure, "value", G_TYPE_BOOLEAN,
                                  *(gboolean *)(p_msg->value), NULL);
                break;
              case G_TYPE_INT:
                gst_structure_set(structure, "value", G_TYPE_INT,
                                  *((gint *)(p_msg->value)), NULL);
                break;
              case G_TYPE_UINT:
                gst_structure_set(structure, "value", G_TYPE_UINT,
                                  *((guint *)(p_msg->value)), NULL);
                break;
              case G_TYPE_LONG:
                gst_structure_set(structure, "value", G_TYPE_LONG,
                                  *((glong *)(p_msg->value)), NULL);
                break;
              case G_TYPE_ULONG:
                gst_structure_set(structure, "value", G_TYPE_ULONG,
                                  *((gulong *)(p_msg->value)), NULL);
                break;
              case G_TYPE_INT64:
                gst_structure_set(structure, "value", G_TYPE_INT64,
                                  *((gint64 *)(p_msg->value)), NULL);
                break;
              case G_TYPE_UINT64:
                gst_structure_set(structure, "value", G_TYPE_UINT64,
                                  *((guint64 *)(p_msg->value)), NULL);
                break;
              case G_TYPE_FLOAT:
                gst_structure_set(structure, "value", G_TYPE_FLOAT,
                                  *((gfloat *)(p_msg->value)), NULL);
                break;
              case G_TYPE_DOUBLE:
                gst_structure_set(structure, "value", G_TYPE_DOUBLE,
                                  *((gdouble *)(p_msg->value)), NULL);
                break;
              case G_TYPE_STRING:
                gst_structure_set(structure, "value", G_TYPE_STRING,
                                  p_msg->value, NULL);
                break;
              default:
                if (g_type_is_a(valtype, G_TYPE_ENUM)) {
                    gst_structure_set(structure, "value", valtype,
                                      *((gint *)(p_msg->value)), NULL);
                }
                else if (g_type_is_a(valtype, G_TYPE_FLAGS)) {
                    gst_structure_set(structure, "value", valtype,
                                      *((guint *)(p_msg->value)), NULL);
                }
                break;
            }
            ls_message(pipeline, config_name, pipeline_name, *msg, structure,
                       NULL);
            g_free(propname);
        }
    }
    g_strfreev(name_parts);
    return NULL;
}

request_cb_fn_t
wcbe_mc_property_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PROPERTY_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_property_msg_t **p_msg = (wcbe_property_msg_t **)msg;
    (*p_msg)->value_len = 0;
    gchar **name_parts = g_strsplit((*p_msg)->property, G_DIR_SEPARATOR_S, 5);
    if (g_strv_length(name_parts) == 4) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        const gchar *property_name = name_parts[3];
        if (strcmp(config_name, pipeline->config->name) == 0
            && strcmp(pipeline_name, wcbe_input_stage_pipeline_name) == 0) {
            // input stage element property
            if (pipeline->is_pipeline != NULL) {
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->is_pipeline), element_name);
                if (element != NULL) {
                    GParamSpec *param = g_object_class_find_property(
                        G_OBJECT_GET_CLASS(G_OBJECT(element)), property_name);
                    if (param != NULL) {
                        void prep_msg(GType type, gsize value_len) {
                            *msgsize = sizeof(wcbe_property_msg_t) + value_len;
                            *p_msg = g_realloc(*p_msg, *msgsize);
                            (*p_msg)->value_len = value_len;
                            g_strlcpy((*p_msg)->type, g_type_name(type),
                                      sizeof((*p_msg)->type));
                        }
                        void get_simple(void) {
                            g_object_get(G_OBJECT(element), property_name,
                                         (*p_msg)->value, NULL);
                        }
                        switch (param->value_type) {
                          case G_TYPE_CHAR:
                            // gint8 is the correct type according to
                            // GObject docs
                            prep_msg(G_TYPE_CHAR, sizeof(gint8));
                            get_simple();
                            break;
                          case G_TYPE_UCHAR:
                            prep_msg(G_TYPE_UCHAR, sizeof(guchar));
                            get_simple();
                            break;
                          case G_TYPE_BOOLEAN:
                            prep_msg(G_TYPE_BOOLEAN, sizeof(gboolean));
                            get_simple();
                            break;
                          case G_TYPE_INT:
                            prep_msg(G_TYPE_INT, sizeof(gint));
                            get_simple();
                            break;
                          case G_TYPE_UINT:
                            prep_msg(G_TYPE_UINT, sizeof(guint));
                            get_simple();
                            break;
                          case G_TYPE_LONG:
                            prep_msg(G_TYPE_LONG, sizeof(glong));
                            get_simple();
                            break;
                          case G_TYPE_ULONG:
                            prep_msg(G_TYPE_ULONG, sizeof(gulong));
                            get_simple();
                            break;
                          case G_TYPE_INT64:
                            prep_msg(G_TYPE_INT64, sizeof(gint64));
                            get_simple();
                            break;
                          case G_TYPE_UINT64:
                            prep_msg(G_TYPE_UINT64, sizeof(guint64));
                            get_simple();
                            break;
                          case G_TYPE_FLOAT:
                            prep_msg(G_TYPE_FLOAT, sizeof(gfloat));
                            get_simple();
                            break;
                          case G_TYPE_DOUBLE:
                            prep_msg(G_TYPE_DOUBLE, sizeof(gdouble));
                            get_simple();
                            break;
                          case G_TYPE_STRING: {
                              gchar *str;
                              g_object_get(G_OBJECT(element), property_name,
                                           &str, NULL);
                              guint len = strlen(str) + 1;
                              prep_msg(G_TYPE_STRING, len);
                              g_strlcpy((*p_msg)->value, str, len);
                              break;
                          }
                          default:
                            if (g_type_is_a(param->value_type, G_TYPE_ENUM)) {
                                prep_msg(param->value_type, sizeof(gint));
                                get_simple();
                            }
                            else if (g_type_is_a(param->value_type,
                                                 G_TYPE_FLAGS)) {
                                prep_msg(param->value_type, sizeof(guint));
                                get_simple();
                            }
                            break;
                        }
                    }
                    gst_object_unref(GST_OBJECT(element));
                }
            }
        }
        else {
            // property of element (perhaps) in lag processing stage
            gchar *propname = g_strjoin(
                G_DIR_SEPARATOR_S, element_name, property_name, NULL);
            GstStructure *structure =
                gst_structure_new(GET_PROPERTY_MSG_NAME,
                                  "property", G_TYPE_STRING, propname,
                                  NULL);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, property_query);
            g_free(propname);
        }
    }
    g_strfreev(name_parts);
    return result;
}

request_cb_fn_t
wcbe_mc_pad_list_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PAD_LIST_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_pad_list_msg_t **p_msg = (wcbe_pad_list_msg_t **)msg;
    (*p_msg)->list_len = 0;
    gchar **name_parts = g_strsplit((*p_msg)->element, G_DIR_SEPARATOR_S, 4);
    if (g_strv_length(name_parts) == 3) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        if (strcmp(config_name, pipeline->config->name) == 0
            && strcmp(pipeline_name, wcbe_input_stage_pipeline_name) == 0) {
            // input stage element pad
            if (pipeline->is_pipeline != NULL) {
                GPtrArray *pads = g_ptr_array_new();
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->is_pipeline), element_name);
                if (element != NULL) {
                    (*p_msg)->list_len = 1;
                    GstIterator *pads_iter = gst_element_iterate_pads(element);
                    void pad_name(gpointer data, gpointer user_data) {
                        GstPad *pad = GST_PAD(data);
                        gchar *name = gst_pad_get_name(pad);
                        gst_object_unref(GST_OBJECT(pad));
                        g_ptr_array_add(pads, name);
                        (*p_msg)->list_len += strlen(name) + 1;

                    }
                    GstIteratorResult iter_result =
                        gst_iterator_foreach(pads_iter, pad_name, NULL);
                    gst_iterator_free(pads_iter);
                    if (iter_result == GST_ITERATOR_ERROR) {
                        guint i;
                        for (i = 0; i < pads->len; ++i) {
                            gchar *name = g_ptr_array_index(pads, i);
                            g_free(name);
                            (*p_msg)->list_len = 1;
                        }
                        pads->len = 0;
                    }
                    gst_object_unref(GST_OBJECT(element));
                }
                *msgsize = sizeof(wcbe_pad_list_msg_t) + (*p_msg)->list_len;
                *p_msg = g_realloc(*p_msg, *msgsize);
                concat_string_list(pads, (*p_msg)->list, TRUE);
                g_ptr_array_free(pads, TRUE);
            }
        }
        else {
            GstStructure *structure =
                gst_structure_new(PAD_LIST_MSG_NAME,
                                  "element", G_TYPE_STRING, element_name,
                                  NULL);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, pad_list_query);
        }
    }
    g_strfreev(name_parts);

    return result;
}

request_cb_fn_t
wcbe_mc_pad_caps_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PAD_CAPS_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_pad_caps_msg_t *p_msg = (wcbe_pad_caps_msg_t *)*msg;
    p_msg->caps[0] = '\0';
    gchar **name_parts = g_strsplit(p_msg->pad, G_DIR_SEPARATOR_S, 5);
    if (g_strv_length(name_parts) == 4) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        const gchar *pad_name = name_parts[3];
        if (strcmp(config_name, pipeline->config->name) == 0
            && strcmp(pipeline_name, wcbe_input_stage_pipeline_name) == 0) {
            // input stage element property
            if (pipeline->is_pipeline != NULL) {
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->is_pipeline), element_name);
                if (element != NULL) {
                    GstPad *pad =
                        gst_element_get_static_pad(element, pad_name);
                    if (pad != NULL) {
                        GstCaps *caps = gst_pad_get_negotiated_caps(pad);
                        if (caps != NULL) {
                            gchar *str = gst_caps_to_string(caps);
                            g_strlcpy(p_msg->caps, str, sizeof(p_msg->caps));
                            g_free(str);
                            gst_caps_unref(caps);
                        }
                        gst_object_unref(GST_OBJECT(pad));
                    }
                    gst_object_unref(GST_OBJECT(element));
                }
            }
        }
        else {
        // element (perhaps) in lag processing stage
        gchar *pname =
            g_strjoin(G_DIR_SEPARATOR_S, element_name, pad_name, NULL);
        GstStructure *structure =
            gst_structure_new(PAD_CAPS_MSG_NAME,
                              "pad", G_TYPE_STRING, pname,
                              NULL);
        g_assert(structure != NULL);
        result = ls_message(pipeline, config_name, pipeline_name, *msg,
                            structure, pad_caps_query);
        g_free(pname);
        }
    }
    g_strfreev(name_parts);
    return result;
}

request_cb_fn_t
wcbe_mc_pad_peer_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize)
{
    g_return_val_if_fail((*msg)->type == WCBE_PAD_PEER_MSG, NULL);

    request_cb_fn_t result = NULL;
    wcbe_pad_peer_msg_t *p_msg = (wcbe_pad_peer_msg_t *)*msg;
    p_msg->peer[0] = '\0';
    gchar **name_parts = g_strsplit(p_msg->pad, G_DIR_SEPARATOR_S, 5);
    if (g_strv_length(name_parts) == 4) {
        const gchar *config_name = name_parts[0];
        const gchar *pipeline_name = name_parts[1];
        const gchar *element_name = name_parts[2];
        const gchar *pad_name = name_parts[3];
        if (strcmp(config_name, pipeline->config->name) == 0
            && strcmp(pipeline_name, wcbe_input_stage_pipeline_name) == 0) {
            // input stage element property
            if (pipeline->is_pipeline != NULL) {
                GstElement *element = gst_bin_get_by_name(
                    GST_BIN(pipeline->is_pipeline), element_name);
                if (element != NULL) {
                    GstPad *pad =
                        gst_element_get_static_pad(element, pad_name);
                    if (pad != NULL)
                        if (gst_pad_is_linked(pad)) {
                            GstPad *peer = gst_pad_get_peer(pad);
                            g_assert(peer != NULL);
                            GstObject *parent = gst_pad_get_parent(peer);
                            g_assert(parent != NULL);
                            g_snprintf(p_msg->peer, sizeof(p_msg->peer),
                                       "%s/%s", GST_OBJECT_NAME(parent),
                                       GST_OBJECT_NAME(peer));
                            gst_object_unref(parent);
                            gst_object_unref(GST_OBJECT(peer));
                        }
                        gst_object_unref(GST_OBJECT(pad));
                    }
                gst_object_unref(GST_OBJECT(element));
            }
        }
        else {
            // element (perhaps) in lag processing stage
            gchar *pname =
                g_strjoin(G_DIR_SEPARATOR_S, element_name, pad_name, NULL);
            GstStructure *structure =
                gst_structure_new(PAD_PEER_MSG_NAME,
                                  "pad", G_TYPE_STRING, pname,
                                  NULL);
            g_assert(structure != NULL);
            result = ls_message(pipeline, config_name, pipeline_name, *msg,
                                structure, pad_peer_query);
            g_free(pname);
        }
    }
    g_strfreev(name_parts);
    return result;
}

static void
concat_string_list(GPtrArray *strings, gchar *result, gboolean free_strings)
{
    guint i;
    for (i = 0; i < strings->len; ++i) {
        gchar *str = g_ptr_array_index(strings, i);
        guint len = strlen(str) + 1;
        memcpy(result, str, len);
        result += len;
        if (free_strings) {
            g_free(str);
        }
    }
    *result = '\0';
}

static request_cb_fn_t
ls_message(pipeline_t *pipeline, const gchar *config_name,
           const gchar *ls_pipeline_name, const wcbe_msg_t *msg,
           GstStructure *structure, request_cb_fn_t cb)
{
    request_cb_fn_t result;
    // elements (perhaps) in a lag processing stage pipeline
    if (cb != NULL) {
        gchar *id = g_strdup_printf("%p", msg);
        gst_structure_set(structure,
                          "id", G_TYPE_STRING, id,
                          NULL);
        g_free(id);
    }
    if (forward_message(pipeline, config_name, ls_pipeline_name, structure)) {
        result = cb;
    }
    else {
        gst_structure_free(structure);
        result = NULL;
    }
    return result;
}

static gboolean
forward_message(pipeline_t *pipeline, const gchar *config_name,
                const gchar *ls_pipeline_name, GstStructure *structure)
{
    gboolean result = FALSE;
    configuration_t *config =
        g_hash_table_lookup(pipeline->configs, config_name);
    if (config != NULL) {
        gchar *full_pipeline_name =
            g_strjoin(G_DIR_SEPARATOR_S, config_name, ls_pipeline_name, NULL);
        ls_pipeline_t *ls_pipeline =
            wcbe_find_ls_pipeline(config->ls_pipelines, full_pipeline_name);
        if (ls_pipeline != NULL) {
            gchar *msg = gst_structure_to_string(structure);
            wcbe_stream_msg_t stream_msg;
            stream_msg.object.type = WCBE_STREAM_MSG;
            stream_msg.structure_len = strlen(msg) + 1;
            struct iovec iovec[2] =
                {{.iov_base = &stream_msg, .iov_len = sizeof(stream_msg)},
                 {.iov_base = msg, .iov_len = stream_msg.structure_len}};
            struct msghdr msghdr =
                {.msg_name = &(ls_pipeline->sockaddr),
                 .msg_namelen = sizeof(ls_pipeline->sockaddr),
                 .msg_control = NULL, .msg_controllen = 0,
                 .msg_iov = iovec, .msg_iovlen = G_N_ELEMENTS(iovec)};
            errno = 0;
            gssize bytes_written = sendmsg(pipeline->lsp_sockfd, &msghdr, 0);
            if (G_LIKELY(bytes_written
                         == sizeof(stream_msg) + stream_msg.structure_len)) {
                result = TRUE;
            }
            else {
                const gchar *errstr;
                if (errno != 0) {
                    errstr = g_strerror(errno);
                }
                else {
                    errstr = "'errno' not set";
                }
                GST_WARNING("Failed to write stream object to socket %s: %s",
                            ((struct sockaddr_un *)msghdr.msg_name)->sun_path,
                            errstr);
            }
            g_free(msg);
        }
        g_free(full_pipeline_name);
    }
    return result;
}


static configuration_t *
find_config(pipeline_t *pipeline, const gchar *name)
{
    configuration_t *result;
    result = g_hash_table_lookup(pipeline->configs, name);
    if (result == NULL) {
        if (pipeline->pending_configs != NULL) {
            g_assert(pipeline->is_pipeline == NULL
                     || pipeline->flow == WCBE_FLOW_OFF);
            GSList *pconf =
                wcbe_find_pending_config(pipeline->pending_configs, name);
            if (pconf != NULL) {
                result = ((pending_config_t *)(pconf->data))->config;
            }
        }
    }
    return result;
}

static void
element_list_query(pipeline_t *pipeline, wcbe_msg_t **msg,
                   gsize *msgsize, GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_ELEMENT_LIST_MSG);

    wcbe_element_list_msg_t **e_msg = (wcbe_element_list_msg_t **)msg;
    (*e_msg)->list_len = 1;
    const gchar *list = gst_structure_get_string(structure, "list");
    gchar **names = g_strsplit(list, ",", 0);
    GPtrArray *elements = g_ptr_array_new();
    gchar **name = names;
    while (*name != NULL) {
        g_ptr_array_add(elements, *name);
        (*e_msg)->list_len += strlen(*name) + 1;
        ++name;
    }
    *msgsize = sizeof(wcbe_element_list_msg_t) + (*e_msg)->list_len;
    *e_msg = g_realloc(*e_msg, *msgsize);
    concat_string_list(elements, (*e_msg)->list, FALSE);
    g_ptr_array_free(elements, TRUE);
    g_strfreev(names);
}

static void
property_list_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
                    GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PROPERTY_LIST_MSG);

    wcbe_property_list_msg_t **p_msg = (wcbe_property_list_msg_t **)msg;
    (*p_msg)->list_len = 1;
    const gchar *list = gst_structure_get_string(structure, "list");
    gchar **names = g_strsplit(list, ",", 0);
    GPtrArray *properties = g_ptr_array_new();
    gchar **name = names;
    while (*name != NULL) {
        g_ptr_array_add(properties, *name);
        (*p_msg)->list_len += strlen(*name) + 1;
        ++name;
    }
    *msgsize = sizeof(wcbe_property_list_msg_t) + (*p_msg)->list_len;
    *p_msg = g_realloc(*p_msg, *msgsize);
    concat_string_list(properties, (*p_msg)->list, FALSE);
    g_ptr_array_free(properties, TRUE);
    g_strfreev(names);
}

static void
property_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
               GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PROPERTY_MSG);

    wcbe_property_msg_t **p_msg = (wcbe_property_msg_t **)msg;
    const GValue *value = gst_structure_get_value(structure, "value");
    if (value != NULL) {
        void prep_msg(GType type, gsize value_len) {
            *msgsize = sizeof(wcbe_property_msg_t) + value_len;
            *p_msg = g_realloc(*p_msg, *msgsize);
            (*p_msg)->value_len = value_len;
            g_strlcpy((*p_msg)->type, g_type_name(type),
                      sizeof((*p_msg)->type));
        }
        void copy_val(gconstpointer v) {
            memcpy((*p_msg)->value, v, (*p_msg)->value_len);
        }
        switch (G_VALUE_TYPE(value)) {
          case G_TYPE_CHAR: {
              // gint8 is the correct type according to GObject docs
              prep_msg(G_TYPE_CHAR, sizeof(gint8));
              gchar v = g_value_get_char(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_UCHAR: {
              prep_msg(G_TYPE_UCHAR, sizeof(guchar));
              guchar v = g_value_get_uchar(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_BOOLEAN: {
              prep_msg(G_TYPE_BOOLEAN, sizeof(gboolean));
              gboolean v = g_value_get_boolean(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_INT: {
              prep_msg(G_TYPE_INT, sizeof(gint));
              gint v = g_value_get_int(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_UINT: {
              prep_msg(G_TYPE_UINT, sizeof(guint));
              guint v = g_value_get_uint(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_LONG: {
              prep_msg(G_TYPE_LONG, sizeof(glong));
              glong v = g_value_get_long(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_ULONG: {
              prep_msg(G_TYPE_ULONG, sizeof(gulong));
              gulong v = g_value_get_ulong(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_INT64: {
              prep_msg(G_TYPE_INT64, sizeof(gint64));
              gint64 v = g_value_get_int64(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_UINT64: {
              prep_msg(G_TYPE_UINT64, sizeof(guint64));
              guint64 v = g_value_get_uint64(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_FLOAT: {
              prep_msg(G_TYPE_FLOAT, sizeof(gfloat));
              gfloat v = g_value_get_float(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_DOUBLE: {
              prep_msg(G_TYPE_DOUBLE, sizeof(gdouble));
              gdouble v = g_value_get_double(value);
              copy_val(&v);
              break;
          }
          case G_TYPE_STRING: {
              const gchar *str = g_value_get_string(value);
              guint len = strlen(str) + 1;
              prep_msg(G_TYPE_STRING, len);
              copy_val(str);
              break;
          }
          default:
            if (g_type_is_a(G_VALUE_TYPE(value), G_TYPE_ENUM)) {
                prep_msg(G_VALUE_TYPE(value), sizeof(gint));
                gint v = g_value_get_enum(value);
                copy_val(&v);
            }
            else if (g_type_is_a(G_VALUE_TYPE(value), G_TYPE_FLAGS)) {
                prep_msg(G_VALUE_TYPE(value), sizeof(guint));
                guint v = g_value_get_flags(value);
                copy_val(&v);
            }
            break;
        }
    }
    else {
        (*p_msg)->value_len = 0;
    }
}

static void
pad_list_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
               GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PAD_LIST_MSG);

    wcbe_pad_list_msg_t **p_msg = (wcbe_pad_list_msg_t **)msg;
    (*p_msg)->list_len = 1;
    const gchar *list = gst_structure_get_string(structure, "list");
    gchar **names = g_strsplit(list, ",", 0);
    GPtrArray *pads = g_ptr_array_new();
    gchar **name = names;
    while (*name != NULL) {
        g_ptr_array_add(pads, *name);
        (*p_msg)->list_len += strlen(*name) + 1;
        ++name;
    }
    *msgsize = sizeof(wcbe_pad_list_msg_t) + (*p_msg)->list_len;
    *p_msg = g_realloc(*p_msg, *msgsize);
    concat_string_list(pads, (*p_msg)->list, FALSE);
    g_ptr_array_free(pads, TRUE);
    g_strfreev(names);
}

static void
pad_caps_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
               GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PAD_CAPS_MSG);

    wcbe_pad_caps_msg_t *p_msg = (wcbe_pad_caps_msg_t *)*msg;
    const gchar *caps = gst_structure_get_string(structure, "caps");
    g_strlcpy(p_msg->caps, caps, sizeof(p_msg->caps));
}

static void
pad_peer_query(pipeline_t *pipeline, wcbe_msg_t **msg, gsize *msgsize,
               GstStructure *structure)
{
    g_return_if_fail((*msg)->type == WCBE_PAD_PEER_MSG);

    wcbe_pad_peer_msg_t *p_msg = (wcbe_pad_peer_msg_t *)*msg;
    const gchar *peer = gst_structure_get_string(structure, "peer");
    g_strlcpy(p_msg->peer, peer, sizeof(p_msg->peer));
}
