/*
 * dump_frame.c
 *
 * Functions for dumping lag frames in ASCII-format file.
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <dump_frame.h>

// this module is not MT-safe

static void init(const char *prefix);

static gboolean initialized = FALSE;
#define NUM_CCCS 64
static char *output_filename_prefix;
static const char *default_prefix = "ccc";
static const char *output_filename_suffix = ".txt";
static FILE *output_files[NUM_CCCS + 1];
static off64_t max_file_size = 0;
static off_t write_lta(const lf_data_frame_word_t *df);
static off_t write_cc(const lf_data_frame_word_t *df);
static off_t write_unknown(const lf_data_frame_word_t *df);

#define BITMASK(width, offset) \
    (((((lf_data_frame_word_t)1) << (width)) - 1) << (offset))
#define BITFIELD_EXTRACT(bitfield, width, offset) \
    (((bitfield) & BITMASK(width, offset)) >> (offset))
#define BITFIELD_PART(value, width, offset) \
    ((((lf_data_frame_word_t)(value)) & \
      ((((lf_data_frame_word_t)1) << (width)) - 1)) << (offset))

/* dump_frame_set_file_prefix
 *
 * Set the prefix of file names. File names have the pattern
 * "<prefix>-<ccc number>.txt".
 */
void
dump_frame_set_file_prefix(const char *prefix)
{
    init(prefix);
}

/* dump_frame
 *
 * Write a lag frame to an ASCII file.
 */
off64_t
dump_frame(const lf_data_frame_word_t *df)
{
    if (!initialized) {
        init(default_prefix);
    }
    off_t fsize;
    switch (LF_DATA_FRAME_FTYPE(df)) {
    case LF_LTA_DATA_FRAME_FTYPE:
        fsize = write_lta(df);
        break;
    case LF_CC_DATA_FRAME_FTYPE:
        fsize = write_cc(df);
        break;
    default:
        fsize = write_unknown(df);
        break;
    }
    g_assert(fsize != -1);
    if (fsize > max_file_size) {
        max_file_size = fsize;
    }
    return max_file_size;
}

/* dump_frame_close_all
 *
 * Close all files to which lag frames are being written.
 */
void
dump_frame_close_all(void)
{
    if (initialized) {
        unsigned i;
        for (i = 0; i <= NUM_CCCS; ++i) {
            if (output_files[i] != NULL) {
                fclose(output_files[i]);
            }
            output_files[i] = NULL;
        }
        max_file_size = 0;
    }
}

off64_t
dump_frame_max_file_size(void)
{
    return max_file_size;
}

/* init
 *
 * Prepare the open file array, and copy the file name prefix.
 */
static void
init(const char *prefix)
{
    if (initialized) {
        dump_frame_close_all();
        g_free(output_filename_prefix);
    }
    output_filename_prefix = g_strdup(prefix);
    unsigned i;
    for (i = 0; i <= NUM_CCCS; ++i) {
        output_files[i] = NULL;
    }
    initialized = TRUE;
}

/* get_output_file
 *
 * Get the file for a given ccc number.
 */
static FILE *
get_output_file(unsigned ccc)
{
    FILE **result = &output_files[ccc];
    if (*result == NULL) {
        char *name =
            g_strdup_printf("%s-%u%s", output_filename_prefix, ccc,
                            output_filename_suffix);
        *result = fopen(name, "w+");
        if (*result == NULL) {
            // abort, since we want to avoid repeatedly trying to open
            // the same file
            g_error("Failed to open output file %d: %s", ccc, strerror(errno));
        }
        g_free(name);
    }
    return *result;
}

/* write_lta
 *
 * Write an LTA lag frame to an ASCII file according to its ccc number.
 */
static off_t
write_lta(const lf_data_frame_word_t *df)
{
    unsigned ccc = BITFIELD_EXTRACT(g_ntohl(df[1]), 4, 3);
    FILE *f = get_output_file(ccc);
    lf_data_frame_word_t word;
    lf_data_frame_word_t checksum = 0;

#   define print_hex_word()  fprintf(f, "%08x    ", word)
#   define next_word() G_STMT_START {checksum += word; ++df; \
                                     word = g_ntohl(*df);} G_STMT_END

    unsigned num_lags = LF_LTA_DATA_FRAME_NUM_LAGS(df);

    // word 0
    word = *df; // this does not need byte order conversion since
                // LF_START_SYNC_WORD is expressed in network byte
                // order
    print_hex_word();
    fprintf(f, "W0: START SYNC WORD -- %s\n",
            ((word == LF_START_SYNC_WORD) ? "OK" : "ERROR"));
    next_word();

    // word 1
    print_hex_word();
    fprintf(f, "W1: Start_BlockY=%u NBlocks=%u nlags=%u ChipID-Y=%u "
            "ChipID-X=%u CCC=%u FType=%u\n",
            BITFIELD_EXTRACT(word, 8, 24), BITFIELD_EXTRACT(word, 8, 16),
            BITFIELD_EXTRACT(word, 3, 13), BITFIELD_EXTRACT(word, 3, 10),
            BITFIELD_EXTRACT(word, 3, 7), BITFIELD_EXTRACT(word, 4, 3),
            BITFIELD_EXTRACT(word, 3, 0));
    next_word();

    // word 2
    print_hex_word();
    fprintf(f, "W2: BBID-Y=%u SBID-Y=%u SID-Y=%u BBID-X=%u SBID-X=%u "
            "SID-X=%u\n",
            BITFIELD_EXTRACT(word, 3, 29), BITFIELD_EXTRACT(word, 5, 24),
            BITFIELD_EXTRACT(word, 8, 16), BITFIELD_EXTRACT(word, 3, 13),
            BITFIELD_EXTRACT(word, 5, 8), BITFIELD_EXTRACT(word, 8, 0));
    next_word();

    // word 3
    print_hex_word();
    fprintf(f, "W3: B31=%u Yin=%u Xin=%u YSyncerr=%u XSyncerr=%u ACC_OV=%u"
            " OVR=%u FRAME_COUNT=%u RECIRC_BLK-Y=%u RECIRC_BLK-X=%u\n",
            BITFIELD_EXTRACT(word, 1, 31), BITFIELD_EXTRACT(word, 1, 30),
            BITFIELD_EXTRACT(word, 1, 29), BITFIELD_EXTRACT(word, 1, 28),
            BITFIELD_EXTRACT(word, 1, 27), BITFIELD_EXTRACT(word, 1, 26),
            BITFIELD_EXTRACT(word, 1, 25), BITFIELD_EXTRACT(word, 9, 16),
            BITFIELD_EXTRACT(word, 8, 8), BITFIELD_EXTRACT(word, 8, 0));
    next_word();

    // word 4
    print_hex_word();
    fprintf(f, "W4: TIMESTAMP-0=%10u\n", word);
    next_word();

    // word 5
    print_hex_word();
    fprintf(f, "W5: TIMESTAMP-1=%10u\n", word);
    next_word();

    // word 6
    print_hex_word();
    fprintf(f, "W6: DVCOUNT-Center=%10u\n", word);
    next_word();

    // word 7
    print_hex_word();
    fprintf(f, "W7: DVCOUNT-Edge  =%10u\n", word);
    next_word();

    // word 8
    print_hex_word();
    fprintf(f, "W8: Board S/N=%u LTA bin=%u\n",
            BITFIELD_EXTRACT(word, 16, 16), BITFIELD_EXTRACT(word, 16, 0));
    next_word();

    // lags
    unsigned i;
    for (i = 0; i < num_lags; ++i) {
        print_hex_word();
        fprintf(f, "Lag%5u =%10u\n", i, word);
        next_word();
        print_hex_word();
        fprintf(f, "Lag%5u =%10u\n", i, word);
        next_word();
    }

    // end sync word
    print_hex_word();
    fprintf(f, "W%u: END SYNC WORD -- %s\n", 2 * num_lags + 9,
            ((word == g_ntohl(LF_END_SYNC_WORD)) ? "OK" : "ERROR"));
    next_word();

    // checksum word
    print_hex_word();
    fprintf(f, "W%u: Checksum calculated %s\n\n", 2 * num_lags + 10,
            ((word == checksum) ? "OK" : "ERROR"));

#   undef next_word
#   undef print_hex_word
    return ftello(f);
}

/* write_cc
 *
 * Write a correlator chip lag frame to an ASCII file according to its
 * ccc number.
 */
static off_t
write_cc(const lf_data_frame_word_t *df)
{
    unsigned ccc = BITFIELD_EXTRACT(g_ntohl(df[1]), 6, 3);
    FILE *f = get_output_file(ccc);
    lf_data_frame_word_t word;
    lf_data_frame_word_t checksum = 0;

#   define print_hex_word()  fprintf(f, "%08x    ", word)
#   define next_word() G_STMT_START {checksum += word; ++df; \
                                     word = g_ntohl(*df);} G_STMT_END

    unsigned num_lags = LF_CC_DATA_FRAME_NUM_LAGS(df);

    // word 0
    word = *df; // this doesn't need byte order conversion since
                // LF_START_SYNC_WORD is expressed in network byte
                // order
    print_hex_word();
    fprintf(f, "W0: START SYNC WORD -- %s\n",
            ((word == LF_START_SYNC_WORD) ? "OK" : "ERROR"));
    next_word();

    // word 1
    print_hex_word();
    fprintf(f, "W3: B31=%u Yin=%u Xin=%u YSyncerr=%u XSyncerr=%u ACC_OV=%u"
            " OVR=%u Rsrv=%u NUM_CLAGS=%u CCC=%u Cmmd=%u\n",
            BITFIELD_EXTRACT(word, 1, 31), BITFIELD_EXTRACT(word, 1, 30),
            BITFIELD_EXTRACT(word, 1, 29), BITFIELD_EXTRACT(word, 1, 28),
            BITFIELD_EXTRACT(word, 1, 27), BITFIELD_EXTRACT(word, 1, 26),
            BITFIELD_EXTRACT(word, 1, 25), BITFIELD_EXTRACT(word, 6, 19),
            BITFIELD_EXTRACT(word, 10, 9), BITFIELD_EXTRACT(word, 6, 3),
            BITFIELD_EXTRACT(word, 3, 0));
    next_word();

    // word 2
    print_hex_word();
    fprintf(f, "W2: BBID-Y=%u SBID-Y=%u SID-Y=%u BBID-X=%u SBID-X=%u "
            "SID-X=%u\n",
            BITFIELD_EXTRACT(word, 3, 29), BITFIELD_EXTRACT(word, 5, 24),
            BITFIELD_EXTRACT(word, 8, 16), BITFIELD_EXTRACT(word, 3, 13),
            BITFIELD_EXTRACT(word, 5, 8), BITFIELD_EXTRACT(word, 8, 0));
    next_word();

    // word 3
    print_hex_word();
    fprintf(f, "W3: LTA/Phase bin=%u RECIRC_BLK-Y=%u RECIRC_BLK-X=%u\n",
            BITFIELD_EXTRACT(word, 16, 16), BITFIELD_EXTRACT(word, 8, 8),
            BITFIELD_EXTRACT(word, 8, 0));
    next_word();

    // word 4
    print_hex_word();
    fprintf(f, "W4: TIMESTAMP-0=%10u\n", word);
    next_word();

    // word 5
    print_hex_word();
    fprintf(f, "W5: TIMESTAMP-1=%10u\n", word);
    next_word();

    // word 6
    print_hex_word();
    fprintf(f, "W6: DVCOUNT-Center=%10u\n", word);
    next_word();

    // word 7
    print_hex_word();
    fprintf(f, "W7: DVCOUNT-Edge  =%10u\n", word);
    next_word();

    // word 8
    print_hex_word();
    fprintf(f, "W8: DATA_BIAS     =%10u\n", word);
    next_word();

    // lags
    unsigned i;
    for (i = 0; i < num_lags; ++i) {
        print_hex_word();
        fprintf(f, "Lag%5u =%10u\n", i, word);
        next_word();
        print_hex_word();
        fprintf(f, "Lag%5u =%10u\n", i, word);
        next_word();
    }

    // end sync word
    print_hex_word();
    fprintf(f, "W%u: END SYNC WORD -- %s\n", 2 * num_lags + 9,
            ((word == g_ntohl(LF_END_SYNC_WORD)) ? "OK" : "ERROR"));
    next_word();

    // checksum word
    print_hex_word();
    fprintf(f, "W%u: Checksum calculated %s\n\n", 2 * num_lags + 10,
            ((word == checksum) ? "OK" : "ERROR"));

#   undef next_word
#   undef print_hex_word
    return ftello(f);
}

/* write_unknown
 *
 * Write a lag frame of unknown FTYPE to an ASCII file. The file
 * follows the same naming pattern as the other output files, but uses
 * an "impossible" ccc value (i.e, the number of cccs)to distinguish
 * itself.
 */
static off_t
write_unknown(const lf_data_frame_word_t *df)
{
    // use an "impossible" ccc value to identify these frames
    FILE *f = get_output_file(NUM_CCCS);

#   define print_hex_word()  fprintf(f, "%08x    ", *df)
#   define next_word() ++df

    //word 0
    print_hex_word();
    fprintf(f, "W0: START SYNC WORD -- %s\n",
            ((*df == LF_START_SYNC_WORD) ? "OK" : "ERROR"));
    next_word();

    // from here on, just print out the word in hex

    unsigned i;
    for (i = 0; i < 8; ++i) {
        print_hex_word();
        fprintf(f, "W%u\n", i + 1);
        next_word();
    }
#   undef next_word
#   undef print_hex_word
    return ftello(f);
}
