#ifndef WCBE_H_
#define WCBE_H_

#include <string.h>
#include <complex.h>
#include <glib.h>
#include <sys/socket.h>
#include <gst/gst.h>
#include <wcbe_lag_frame.h>

G_BEGIN_DECLS

// TODO: the following definition should come from "config.h"
#define TERM_VARRAY(type, arr) type arr[]

#define WCBE_RAW_MIME "application/x-widar-raw"
#define WCBE_MIME "application/x-widar"
#define WCBE_LAGSET_MIME "application/x-widar-lagset"
#define WCBE_LAG_SPECTRA_MIME "application/x-widar-lag-spectra"
#define WCBE_LAG_FREQUENCY_MIME "application/x-widar-frequency-spectra"

#define WIDAR_CLOCK_RATE 128000000

#define TIMEVALUE(ts) ((ts)->seconds * (guint64)WIDAR_CLOCK_RATE + \
                       (ts)->clock_cycles)

typedef struct header_t {
    lf_data_frame_word_t key[LF_KEY_LENGTH];
    guint lag_offset;
    guint16 num_lags;
    guint8 status;
    lf_widar_time_t timestamp;
    guint32 data_valid_center;
    guint32 data_valid_edge;
    union {
        guint16 frame_count;
        guint32 data_bias;
    };
} header_t;

#define HEADER_START_BLOCKY(h) (LF_BITFIELD_EXTRACT((h)->key[0], 8, 24))
#define HEADER_NBLOCKS(h) (LF_BITFIELD_EXTRACT((h)->key[0], 8, 16))
#define HEADER_CHIPID(h) (LF_BITFIELD_EXTRACT((h)->key[0], 6, 7))
#define HEADER_CCC(h) (LF_BITFIELD_EXTRACT((h)->key[0], 4, 3))
#define HEADER_FTYPE(h) (LF_BITFIELD_EXTRACT((h)->key[0], 3, 0))

#define HEADER_BBID_Y(h) (LF_BITFIELD_EXTRACT((h)->key[1], 3, 29))
#define HEADER_SBID_Y(h) (LF_BITFIELD_EXTRACT((h)->key[1], 5, 24))
#define HEADER_SID_Y(h) (LF_BITFIELD_EXTRACT((h)->key[1], 8, 16))
#define HEADER_BBID_X(h) (LF_BITFIELD_EXTRACT((h)->key[1], 3, 13))
#define HEADER_SBID_X(h) (LF_BITFIELD_EXTRACT((h)->key[1], 5, 8))
#define HEADER_SID_X(h) (LF_BITFIELD_EXTRACT((h)->key[1], 8, 0))

#define HEADER_RECIRC_BLK_Y(h) (LF_BITFIELD_EXTRACT((h)->key[2], 8, 8))
#define HEADER_RECIRC_BLK_X(h) (LF_BITFIELD_EXTRACT((h)->key[2], 8, 0))

#define HEADER_BOARD_SN(h) (LF_BITFIELD_EXTRACT((h)->key[3], 16, 16))
#define HEADER_LTA_BIN(h) (LF_BITFIELD_EXTRACT((h)->key[3], 16, 0))

#define MAX_LAGS_PER_FRAME 128
#define MAX_LAGSET_PRODUCT_ID_LEN 80

#define WCBE_EVENT_CREATE_CONFIG \
  GST_EVENT_MAKE_TYPE(71, GST_EVENT_TYPE_DOWNSTREAM \
                          | GST_EVENT_TYPE_SERIALIZED)
#define WCBE_EVENT_DESTROY_CONFIG \
  GST_EVENT_MAKE_TYPE(72, GST_EVENT_TYPE_DOWNSTREAM \
                          | GST_EVENT_TYPE_SERIALIZED)
#define WCBE_EVENT_ACTIVATE_CONFIG \
  GST_EVENT_MAKE_TYPE(73, GST_EVENT_TYPE_DOWNSTREAM \
                          | GST_EVENT_TYPE_SERIALIZED)

typedef struct lagset_t {
    GQuark id;
    lf_widar_time_t timestamp;
    guint num_frames; // number of frames in a set
    guint num_lags; // number of lags in a set
    guint num_bins; // number of bins (multiple of lag sets)
    guint nbits;
    header_t *headers;
    lf_lag_t *lags;
    TERM_VARRAY(guint8, data);
} lagset_t;
#define SIZEOF_LAGSET_T(nframe, nlag, nbin, nbits) \
  (sizeof(lagset_t) + ((nframe) * (nbin)) * sizeof(header_t) \
   + ((nlag) * (nbin) * (((nbits) == 4) ? 1 : 4)) * sizeof(lf_lag_t))

static inline void lagset_init(lagset_t *ls, guint num_frames, guint num_lags,
                               guint num_bins, guint nbits)
{
    ls->num_frames = num_frames;
    ls->num_lags = num_lags;
    ls->num_bins = num_bins;
    ls->nbits = nbits;
    ls->headers = (header_t *)ls->data;
    ls->lags = (lf_lag_t *)&(ls->headers[num_bins * num_frames]);
    memset(ls->headers, 0, sizeof(header_t) * ls->num_frames * ls->num_bins);
}

typedef struct wcbe_stream_object {
    enum {WCBE_STREAM_LAGSET, WCBE_STREAM_EVENT, WCBE_STREAM_MSG} type;
} wcbe_stream_object_t;

typedef struct wcbe_stream_event {
    wcbe_stream_object_t object;
    GstEventType event;
} wcbe_stream_event_t;

typedef struct wcbe_stream_lagset {
    wcbe_stream_object_t object;
    gchar product_id[MAX_LAGSET_PRODUCT_ID_LEN];
    lagset_t lagset;
} wcbe_stream_lagset_t;

typedef struct wcbe_stream_msg {
    wcbe_stream_object_t object;
    guint structure_len;
    gchar structure[0];
} wcbe_stream_msg_t;

#define ELEMENT_LIST_MSG_NAME "element-list"
#define PROPERTY_LIST_MSG_NAME "property-list"
#define GET_PROPERTY_MSG_NAME "get-property"
#define SET_PROPERTY_MSG_NAME "set-property"
#define PAD_LIST_MSG_NAME "pad-list"
#define PAD_CAPS_MSG_NAME "pad-caps"
#define PAD_PEER_MSG_NAME "pad-peer"

static inline void wcbe_stream_lagset_init_iovec(
    wcbe_stream_lagset_t *object, lagset_t *lagset, gsize lagset_size,
    struct iovec iovec[2])
{
    g_assert(offsetof(wcbe_stream_lagset_t, object) == 0);
    iovec[0].iov_base = object;
    iovec[0].iov_len = offsetof(wcbe_stream_lagset_t, lagset);
    iovec[1].iov_base = lagset;
    iovec[1].iov_len = lagset_size;
}

typedef struct lag_spectra_t {
    guint num_lags;
    guint num_bins;
    guint64 *timestamps;
    float complex *lags;
    float *weights;
    TERM_VARRAY(guint8, data);
} lag_spectra_t;
#define SIZEOF_LAG_SPECTRA_T(nlag, nbin) \
  (sizeof(lag_spectra_t) + (nbin) * \
   ((nlag) * (sizeof(float complex) + sizeof(float)) + sizeof(guint64)))

static inline void lag_spectra_init(lag_spectra_t *spectra, guint num_lags,
                                    guint num_bins)
{
    memset(spectra, 0, SIZEOF_LAG_SPECTRA_T(num_lags, num_bins));
    spectra->num_bins = num_bins;
    spectra->num_lags = num_lags;
    spectra->timestamps = (guint64 *)spectra->data;
    spectra->lags = (float complex *)&(spectra->timestamps[num_bins]);
    spectra->weights = (float *)&(spectra->lags[num_bins * num_lags]);
}

G_END_DECLS

#endif /* WCBE_H_ */
