#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <string.h>

#include <wcbe_decode.h>
#include <wcbe_lag_frame.h>
#include <wcbe.h>

// This will appear in 0.10.13, but it's useful to me now.
//#define GST_BASE_TRANSFORM_FLOW_DROPPED   GST_FLOW_CUSTOM_SUCCESS

GST_DEBUG_CATEGORY_STATIC(wcbe_decode_debug);
#define GST_CAT_DEFAULT wcbe_decode_debug

/* Filter signals and args */
enum
{
    LAST_SIGNAL
};

enum
{
    ARG_0
};

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_RAW_MIME));

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_MIME));

GST_BOILERPLATE(WcbeDecode, wcbe_decode, GstBaseTransform,
                GST_TYPE_BASE_TRANSFORM);

static void wcbe_decode_set_property(GObject *object, guint prop_id,
                                     const GValue *value, GParamSpec *pspec);
static void wcbe_decode_get_property(GObject *object, guint prop_id,
                                     GValue *value, GParamSpec *pspec);

static gboolean wcbe_decode_get_unit_size(GstBaseTransform *trans,
                                          GstCaps *caps, guint *size);
static GstFlowReturn wcbe_decode_transform(GstBaseTransform *trans,
                                           GstBuffer *inbuf,
                                           GstBuffer *outbuf);
static GstCaps *wcbe_decode_transform_caps(GstBaseTransform *trans,
                                           GstPadDirection direction,
                                           GstCaps *caps);

static void
wcbe_decode_base_init(gpointer gclass)
{
    GST_DEBUG_CATEGORY_INIT(wcbe_decode_debug, "wcbe_decode_debug",
                            0, "wcbe_decode debug");
    static GstElementDetails element_details = {
        "WIDAR decoder",
        "Decoder/WIDAR",
        "Decode WIDAR binary lag frames",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get (&src_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get (&sink_factory));
    gst_element_class_set_details (element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_decode_class_init(WcbeDecodeClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GstBaseTransformClass *gstbasetrans_class =
        GST_BASE_TRANSFORM_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_decode_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_decode_get_property);

    gstbasetrans_class->get_unit_size =
        GST_DEBUG_FUNCPTR(wcbe_decode_get_unit_size);
    gstbasetrans_class->transform =
        GST_DEBUG_FUNCPTR(wcbe_decode_transform);
    gstbasetrans_class->transform_caps =
        GST_DEBUG_FUNCPTR(wcbe_decode_transform_caps);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_decode_init(WcbeDecode *filter, WcbeDecodeClass *gclass)
{
}

static void
wcbe_decode_set_property(GObject *object, guint prop_id,
                         const GValue *value, GParamSpec *pspec)
{
    switch (prop_id) {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
wcbe_decode_get_property(GObject *object, guint prop_id,
                         GValue *value, GParamSpec *pspec)
{
    switch (prop_id) {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

/* GstBaseTransform vmethod implementations */

static gboolean
wcbe_decode_get_unit_size(GstBaseTransform *trans, GstCaps *caps, guint *size)
{
    gboolean result;

    GstStructure *structure = gst_caps_get_structure(caps, 0);
    if (gst_structure_has_name(structure, WCBE_MIME)) {
        result = TRUE;
        *size = MAX(sizeof(lf_cc_lag_frame_t), sizeof(lf_lta_lag_frame_t));
    }
    else if (gst_structure_has_name(structure, WCBE_RAW_MIME)) {
        result = TRUE;
        *size = LF_DATA_FRAME_NUM_WORDS(LF_DATA_FRAME_MAX_LAGS)
            * sizeof(lf_data_frame_word_t);
    }
    else {
        result = FALSE;
    }
    return result;
}

static GstFlowReturn
wcbe_decode_transform(GstBaseTransform *trans,
                      GstBuffer *inbuf, GstBuffer *outbuf)
{
    GstFlowReturn result;
    lf_data_frame_word_t *df = (lf_data_frame_word_t *)(inbuf->data);
    switch (LF_DATA_FRAME_FTYPE(df)) {
      case LF_CC_DATA_FRAME_FTYPE: {
          lf_cc_lag_frame_t *lf = (lf_cc_lag_frame_t *)(outbuf->data);
          int rc = lf_to_cc_lag_frame(lf, LF_DATA_FRAME_MAX_LAGS, df);
          if (G_LIKELY(rc == LF_SUCCESS)) {
              result = GST_FLOW_OK;
          }
          else {
              GST_INFO("Failed to decode correlator chip data frame: %s",
                       lf_strerror(rc));
              result = GST_BASE_TRANSFORM_FLOW_DROPPED;
          }
          break;
      }
      case LF_LTA_DATA_FRAME_FTYPE: {
          lf_lta_lag_frame_t *lf = (lf_lta_lag_frame_t *)(outbuf->data);
          int rc = lf_to_lta_lag_frame(lf, LF_DATA_FRAME_MAX_LAGS, df);
          if (G_LIKELY(rc == LF_SUCCESS)) {
              result = GST_FLOW_OK;
          }
          else {
              GST_INFO("Failed to decode LTA data frame: %s",
                       lf_strerror(rc));
              result = GST_BASE_TRANSFORM_FLOW_DROPPED;
          }
          break;
      }
      default:
        result = GST_BASE_TRANSFORM_FLOW_DROPPED;
        break;
    }
    return result;
}

static GstCaps *
wcbe_decode_transform_caps(GstBaseTransform *trans,
                           GstPadDirection direction,
                           GstCaps *caps)
{
    GstStructure *structure = gst_caps_get_structure(caps, 0);
    const gchar *name;
    switch (direction) {
      case GST_PAD_UNKNOWN:
        if (gst_structure_has_name(structure, WCBE_MIME)) {
            name = WCBE_RAW_MIME;
        }
        else if (gst_structure_has_name(structure, WCBE_RAW_MIME)) {
            name = WCBE_MIME;
        }
        else {
            name = NULL;
        }
        break;

      case GST_PAD_SRC:
        if (gst_structure_has_name(structure, WCBE_MIME)) {
            name = WCBE_RAW_MIME;
        }
        else {
            name = NULL;
        }
        break;

      case GST_PAD_SINK:
        if (gst_structure_has_name(structure, WCBE_RAW_MIME)) {
            name = WCBE_MIME;
        }
        else {
            name = NULL;
        }
        break;

      default:
        g_assert_not_reached();
        break;
    }

    GstCaps *result;
    if (name != NULL) {
        result = gst_caps_new_simple(name, NULL);
    }
    else {
        result = gst_caps_new_empty();
    }
    return result;
}
