#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>

#include <wcbe_src.h>
#include <wcbe.h>
#include <wcbe_lag_frame.h>

// The control socket stuff initially was copied from GStreamer's
// 'gstfdsrc' module.
//
//  * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
//  *                    2000 Wim Taymans <wtay@chello.be>
//  *                    2005 Philippe Khalaf <burger@speedy.org>
//
// Licensed under the LGPL v2.

/* the select call is also performed on the control sockets, that way
 * we can send special commands to unblock the select call */
#define CONTROL_STOP            'S'     /* stop the select call */
#define CONTROL_SOCKETS(src)   src->control_sock
#define WRITE_SOCKET(src)      src->control_sock[1]
#define READ_SOCKET(src)       src->control_sock[0]

#define SEND_COMMAND(src, command)          \
G_STMT_START {                              \
  unsigned char c; c = command;             \
  write (WRITE_SOCKET(src), &c, 1);         \
} G_STMT_END

#define READ_COMMAND(src, command, res)        \
G_STMT_START {                                 \
  res = read(READ_SOCKET(src), &command, 1);   \
} G_STMT_END

GST_DEBUG_CATEGORY_STATIC(wcbe_src_debug);
#define GST_CAT_DEFAULT wcbe_src_debug

/* Filter signals and args */
enum
{
    /* FILL ME */
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_URI,
    ARG_WIDAR_MAX_FRAMESIZE,
    ARG_NUM_FRAMES,
    ARG_FPS,
    ARG_LIVE_TIMESTAMPS
};

#define DEFAULT_FPS 100

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_RAW_MIME));

static void wcbe_src_uri_handler_init(gpointer g_iface, gpointer iface_data);

static void
_do_init(GType wcbe_src_type)
{
    static const GInterfaceInfo urihandler_info = {
        wcbe_src_uri_handler_init,
        NULL,
        NULL
    };
    g_type_add_interface_static(wcbe_src_type, GST_TYPE_URI_HANDLER,
                                &urihandler_info);
    GST_DEBUG_CATEGORY_INIT(wcbe_src_debug, "wcbe_src_debug", 0,
                            "wcbe_src debug");
}

GST_BOILERPLATE_FULL(WcbeSrc, wcbe_src, GstPushSrc, GST_TYPE_PUSH_SRC,
                     _do_init);

static void wcbe_src_finalize(GObject *object);
static void wcbe_src_set_property(GObject *object, guint prop_id,
                                  const GValue *value, GParamSpec *pspec);
static void wcbe_src_get_property(GObject *object, guint prop_id,
                                  GValue *value, GParamSpec *pspec);

static gboolean wcbe_src_start(GstBaseSrc *src);
static gboolean wcbe_src_stop(GstBaseSrc *src);
static void wcbe_src_get_times(GstBaseSrc *src, GstBuffer *buffer,
                               GstClockTime *start, GstClockTime *end);
static GstFlowReturn wcbe_src_create(GstPushSrc *src, GstBuffer **buf);
static gboolean wcbe_src_send_event(GstElement *element, GstEvent *event);
static gboolean wcbe_src_unlock(GstBaseSrc *src);
#ifdef BASE_SRC_HAS_UNLOCK_STOP
static gboolean wcbe_src_unlock_stop(GstBaseSrc *src);
#endif

static GstURIType wcbe_src_uri_get_type(void);
static gchar **wcbe_src_uri_get_protocols(void);
static const gchar *wcbe_src_uri_get_uri(GstURIHandler *handler);
static gboolean wcbe_src_uri_set_uri(GstURIHandler *handler, const gchar *uri);

static void
wcbe_src_base_init(gpointer gclass)
{
    static GstElementDetails element_details = {
        "WIDAR frame source",
        "Source",
        "WIDAR frame source",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get (&src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_src_class_init(WcbeSrcClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GstElementClass *gstelement_class = GST_ELEMENT_CLASS(klass);
    GstBaseSrcClass *gstbasesrc_class = GST_BASE_SRC_CLASS(klass);
    GstPushSrcClass *gstpushsrc_class = GST_PUSH_SRC_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_src_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_src_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_src_finalize);

    g_object_class_install_property(
        gobject_class, ARG_LIVE_TIMESTAMPS,
        g_param_spec_boolean ("live-timestamps", "Generate timestamps",
                              "Use live timestamps in data frames?",
                              FALSE, G_PARAM_READWRITE));

    g_object_class_install_property(
        gobject_class, ARG_URI,
        g_param_spec_string("uri", "WIDAR frame source URI",
                            "Set frame source",
                            NULL, G_PARAM_READWRITE));

    g_object_class_install_property(
        gobject_class, ARG_WIDAR_MAX_FRAMESIZE,
        g_param_spec_ulong("widar-max-framesize",
                           "Maximum size of WIDAR data frame",
                           "Set maximum frame size",
                           0, G_MAXULONG,
                           LF_DATA_FRAME_NUM_WORDS(LF_DATA_FRAME_MAX_LAGS)
                           * sizeof(lf_data_frame_word_t),
                           G_PARAM_READWRITE));
    g_object_class_override_property(gobject_class, ARG_WIDAR_MAX_FRAMESIZE,
                                     "blocksize");
    g_object_class_install_property(
        gobject_class, ARG_NUM_FRAMES,
        g_param_spec_uint64("num-frames",
                            "Number of WIDAR data frames sent",
                            "Get number of frames sent",
                            0, G_MAXUINT64,0, G_PARAM_READABLE));

    g_object_class_install_property(
        gobject_class, ARG_FPS,
        g_param_spec_uint("fps",
                          "Frames per second",
                          "Set playback rate of file input",
                          0, G_MAXUINT,
                          DEFAULT_FPS, G_PARAM_READWRITE));

    gstelement_class->send_event = GST_DEBUG_FUNCPTR(wcbe_src_send_event);

    gstbasesrc_class->start = GST_DEBUG_FUNCPTR(wcbe_src_start);
    gstbasesrc_class->stop = GST_DEBUG_FUNCPTR(wcbe_src_stop);
    gstbasesrc_class->get_times = GST_DEBUG_FUNCPTR(wcbe_src_get_times);
    gstbasesrc_class->unlock = GST_DEBUG_FUNCPTR(wcbe_src_unlock);
#ifdef BASE_SRC_HAS_UNLOCK_STOP
    gstbasesrc_class->unlock_stop = GST_DEBUG_FUNCPTR(wcbe_src_unlock_stop);
#endif

    gstpushsrc_class->create = GST_DEBUG_FUNCPTR(wcbe_src_create);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_src_init(WcbeSrc *src, WcbeSrcClass *gclass)
{
    GstBaseSrc *gstbasesrc = GST_BASE_SRC(src);

    gst_base_src_set_format(gstbasesrc, GST_FORMAT_BUFFERS);
    src->uri = NULL;
    src->fps = DEFAULT_FPS;
    src->channel = NULL;
    src->live_timestamps = FALSE;
    src->widar_max_framesize = LF_DATA_FRAME_NUM_WORDS(LF_DATA_FRAME_MAX_LAGS)
        * sizeof(lf_data_frame_word_t);
    READ_SOCKET(src) = 0;
    WRITE_SOCKET(src) = 0;
    src->caps = NULL;
    src->num_frames = 0;
}

static void
wcbe_src_finalize(GObject *object)
{
    WcbeSrc *src = WCBE_SRC(object);
    if (READ_SOCKET(src) > 0) {
        close(READ_SOCKET(src));
    }
    if (WRITE_SOCKET(src) > 0) {
        close(WRITE_SOCKET(src));
    }
    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
wcbe_src_set_property(GObject *object, guint prop_id,
                      const GValue *value, GParamSpec *pspec)
{
    WcbeSrc *src = WCBE_SRC(object);

    switch (prop_id) {
    case ARG_URI:
        gst_uri_handler_set_uri(GST_URI_HANDLER(object),
                                g_value_get_string(value));
        break;
    case ARG_WIDAR_MAX_FRAMESIZE:
        src->widar_max_framesize = g_value_get_ulong(value);
        break;
    case ARG_FPS:
        src->fps = g_value_get_uint(value);
        break;
    case ARG_LIVE_TIMESTAMPS:
        src->live_timestamps = g_value_get_boolean(value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
wcbe_src_get_property(GObject *object, guint prop_id,
                      GValue *value, GParamSpec *pspec)
{
    WcbeSrc *src = WCBE_SRC(object);

    switch (prop_id) {
    case ARG_URI:
        g_value_set_string(value,
                           gst_uri_handler_get_uri(GST_URI_HANDLER(object)));
        break;
    case ARG_WIDAR_MAX_FRAMESIZE:
        g_value_set_ulong(value, src->widar_max_framesize);
        break;
    case ARG_NUM_FRAMES:
#if SIZEOF_UINT64_T == SIZEOF_UNSIGNED
        g_value_set_uint64(value,
                           (guint64)g_atomic_int_get(&(src->num_frames)));
#else
        // The following is exposed to a race condition w.r.t. the
        // pipeline thread updating the num_frames value --- we'll
        // live with it, a mutex lock seems like overkill.
        g_value_set_uint64(value, src->num_frames);
#endif
        break;
    case ARG_FPS:
        g_value_set_uint(value, src->fps);
        break;
    case ARG_LIVE_TIMESTAMPS:
        g_value_set_boolean(value, src->live_timestamps);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

/* GstBaseSrc vmethod implementations */

static gboolean
wcbe_src_start(GstBaseSrc *gstbasesrc)
{
    gboolean result = FALSE;
    WcbeSrc *src = WCBE_SRC(gstbasesrc);

    g_assert(src->channel == NULL);

    if (src->uri == NULL) {
        return FALSE;
    }

    // Create a GIOChannel from which to read frames.
    gchar *location = gst_uri_get_location(src->uri);
    if (src->uri[0] == 'w') { // 'widar' URI protocol
        gchar *colon = strchr(location, ':');
        if (colon != NULL) {
            gulong port = strtoul(&colon[1], NULL, 10);
            if (port != 0) {
                *colon = '\0';
                struct hostent *hostent = gethostbyname(location);
                if (hostent != NULL && hostent->h_addrtype == AF_INET) {
                    gint sock = socket(PF_INET, SOCK_DGRAM, 0);
                    if (sock != -1) {
                        struct sockaddr_in sockaddr =
                            {.sin_family = AF_INET,
                             .sin_port = htons((short)port),
                             .sin_addr = *(struct in_addr*)hostent->h_addr
                            };
                        int rc = bind(sock, (struct sockaddr *)&sockaddr,
                                      sizeof(sockaddr));
                        if (rc != -1) {
                            src->channel = g_io_channel_unix_new(sock);
                            if (src->channel == NULL) {
                                GST_WARNING("Failed to create GIOChannel "
                                            "on socket");
                            }
                        }
                        else {
                            GST_DEBUG("Failed to bind listening socket: %s",
                                      g_strerror(errno));
                        }
                        if (src->channel == NULL) {
                            close(sock);
                        }
                    }
                    else {
                        GST_INFO("Failed to open listening socket");
                    }
                }
                else {
                    GST_INFO("'gethostbyname(%s)' call failed", location);
                }
            }
            else {
                GST_INFO("Failed to parse URI port number");
            }
        }
        else {
            GST_INFO("widar URI has no port number");
        }
    }

    else if (src->uri[0] == 'f') { // 'file' URI protocol
        gint fd = open(location, O_RDONLY);
        if (fd != -1) {
            src->channel = g_io_channel_unix_new(fd);
            if (src->channel == NULL) {
                GST_WARNING("Failed to create GIOChannel on socket");
            }
        }
        else {
            GST_INFO("Failed to open input file '%s': %s", location,
                     g_strerror(errno));
        }
    }
    g_free(location);

    // Initialize the GIOChannel
    if (src->channel != NULL) {
        g_io_channel_set_close_on_unref(src->channel, TRUE);
        GError *err = NULL;
        if ((g_io_channel_set_encoding(src->channel, NULL, &err)
             == G_IO_STATUS_NORMAL)
            && (g_io_channel_set_flags(src->channel, G_IO_FLAG_NONBLOCK, &err)
                == G_IO_STATUS_NORMAL)) {
            g_io_channel_set_buffered(src->channel, FALSE);
        }
        else {
            GST_WARNING("Failed to initialize GIOChannel");
            err = NULL;
            g_io_channel_shutdown(src->channel, FALSE, &err);
            g_io_channel_unref(src->channel);
            src->channel = NULL;
        }
    }

    // Open the control socket pair
    if (src->channel != NULL) {
        gint control_sock[2];
        int rc = socketpair (PF_UNIX, SOCK_STREAM, 0, control_sock);
        if (rc == 0) {
            READ_SOCKET(src) = control_sock[0];
            WRITE_SOCKET(src) = control_sock[1];
            fcntl(READ_SOCKET(src), F_SETFL, O_NONBLOCK);
            fcntl(WRITE_SOCKET(src), F_SETFL, O_NONBLOCK);
            result = TRUE;
        }
        else {
            GST_WARNING("Failed to create control socket pair");
            GError *err = NULL;
            g_io_channel_shutdown(src->channel, FALSE, &err);
            g_io_channel_unref(src->channel);
            src->channel = NULL;
        }
    }

    if (result) {
        src->caps = gst_caps_new_simple(WCBE_RAW_MIME, NULL);
        src->num_frames = 0;
        gst_base_src_set_live(gstbasesrc, TRUE);
        //gst_element_set_state(GST_ELEMENT(src), GST_STATE_PLAYING);
    }

    return result;
}

static gboolean
wcbe_src_stop(GstBaseSrc *gstbasesrc)
{
    WcbeSrc *src = WCBE_SRC(gstbasesrc);
    if (src->channel != NULL) {
        GError *err = NULL;
        g_io_channel_shutdown(src->channel, FALSE, &err);
        g_io_channel_unref(src->channel);
        src->channel = NULL;
    }
    if (READ_SOCKET(src) > 0) {
        close(READ_SOCKET(src));
        READ_SOCKET(src) = 0;
    }
    if (WRITE_SOCKET(src) > 0) {
        close(WRITE_SOCKET(src));
        WRITE_SOCKET(src) = 0;
    }
    if (src->caps != NULL) {
        gst_caps_unref(src->caps);
        src->caps = NULL;
    }
    GST_INFO("Pushed %"G_GUINT64_FORMAT" frames to source pad",
             src->num_frames);
    return TRUE;
}

static gboolean
wcbe_src_unlock(GstBaseSrc *basesrc)
{
    WcbeSrc *src = WCBE_SRC(basesrc);
    SEND_COMMAND(src, CONTROL_STOP);
    return TRUE;
}

#ifdef BASE_SRC_HAS_UNLOCK_STOP
static gboolean
wcbe_src_unlock_stop(GstBaseSrc *basesrc)
{
    WcbeSrc *src = WCBE_SRC(basesrc);

    /* read all stop commands */
    int res = 0;
    while (res >= 0) {
        gchar command;
        READ_COMMAND(src, command, res);
    }
    return TRUE;
}
#endif

static void
wcbe_src_get_times(GstBaseSrc *basesrc, GstBuffer *buffer,
                   GstClockTime *start, GstClockTime *end)
{
    WcbeSrc *src = WCBE_SRC(basesrc);
    if (src->uri[0] == 'f') {
        *start = (src->num_frames * GST_SECOND) / src->fps;
        *end = ((src->num_frames + 1) * GST_SECOND) / src->fps;
    }
}

/* GstPushSrc vmethod implementations */

static GstFlowReturn
wcbe_src_create(GstPushSrc *pushsrc, GstBuffer **outbuf)
{
    GstFlowReturn result;
    GstBaseSrc *basesrc = GST_BASE_SRC(pushsrc);
    WcbeSrc *src = WCBE_SRC(pushsrc);

    // First 'select' on channel fd and control_sock to find which has
    // data to read.

    fd_set readfds;
    gint rc;
    FD_ZERO(&readfds);
    FD_SET(g_io_channel_unix_get_fd(src->channel), &readfds);
    FD_SET(READ_SOCKET(src), &readfds);
    do {
        rc = select(FD_SETSIZE, &readfds, NULL, NULL, NULL);
    } while (rc == -1 && errno == EINTR);
    if (rc == -1) {
        GST_DEBUG("'select' failed: %s", g_strerror(errno));
        return GST_FLOW_ERROR;
    }
    // Quit, if control socket has data.
    if (FD_ISSET(READ_SOCKET(src), &readfds)) {
        GST_DEBUG("stopped");
        return GST_FLOW_WRONG_STATE;
    }

    // Allocate buffer
    guint blocksize = src->widar_max_framesize;
    GstBuffer *buf = NULL;
    GstFlowReturn ret = gst_pad_alloc_buffer(GST_BASE_SRC_PAD(basesrc), 0,
                                             blocksize, src->caps, &buf);
    if (ret != GST_FLOW_OK) {
        return ret;
    }

    // Read a frame
    gsize bytes_read;
    GIOStatus st;
    GError *err = NULL;
    do {
        st = g_io_channel_read_chars(
            src->channel, (gchar *)GST_BUFFER_DATA(buf),
            blocksize, &bytes_read, &err);
    } while (G_UNLIKELY(st == G_IO_STATUS_AGAIN));
    GST_LOG("Read frame from channel");

    // Check read result
    switch (st) {
    case G_IO_STATUS_NORMAL:
        result = GST_FLOW_OK;
        *outbuf = buf;
        GST_BUFFER_SIZE(buf) = bytes_read;
        GST_BUFFER_TIMESTAMP(buf) = GST_CLOCK_TIME_NONE;
        GST_BUFFER_OFFSET(buf) = src->num_frames;
#if SIZEOF_UINT64_T == SIZEOF_UNSIGNED
        g_atomic_int_inc(&(src->num_frames));
#else
        src->num_frames++;
#endif
        if (src->live_timestamps) {
            GTimeVal current;
            g_get_current_time(&current);
            lf_data_frame_word_t *df =
                (lf_data_frame_word_t *)GST_BUFFER_DATA(buf);
            df[4] = g_htonl(current.tv_sec);
            df[5] = g_htonl(current.tv_usec
                            * (WIDAR_CLOCK_RATE / G_USEC_PER_SEC));
            guint num_words =
                ((LF_DATA_FRAME_FTYPE(df) == LF_LTA_DATA_FRAME_FTYPE) ?
                 LF_DATA_FRAME_NUM_WORDS(LF_LTA_DATA_FRAME_NUM_LAGS(df)) :
                 LF_DATA_FRAME_NUM_WORDS(LF_CC_DATA_FRAME_NUM_LAGS(df)))
                - 1;
            lf_data_frame_word_t checksum = 0;
            guint i;
            for (i = 0; i < num_words; ++i) {
                checksum += g_ntohl(df[i]);
            }
            df[i] = g_htonl(checksum);
        }
        break;

    case G_IO_STATUS_EOF:
        GST_DEBUG("Read 0 bytes: EOS");
        gst_buffer_unref(buf);
        result = GST_FLOW_UNEXPECTED;
        break;

    case G_IO_STATUS_ERROR:
        GST_WARNING("Read error on frame source: %s", err->message);
        gst_buffer_unref(buf);
        result = GST_FLOW_ERROR;
        break;

    default:
        g_assert_not_reached();
        break;
    }
    return result;
}

static gboolean
wcbe_src_send_event(GstElement *element, GstEvent *event)
{
    GstBaseSrc *src;
    gboolean result = FALSE;

    src = GST_BASE_SRC (element);

    switch (GST_EVENT_TYPE (event)) {
    case WCBE_EVENT_CREATE_CONFIG:
    case WCBE_EVENT_ACTIVATE_CONFIG:
    case WCBE_EVENT_DESTROY_CONFIG:
        gst_pad_push_event(GST_BASE_SRC_PAD(src), event);
        result = TRUE;
        break;
    default:
        // All events in gst_base_src_send_event have no effect,
        // but EOS handling may change at some time.
        gst_event_unref (event);
        break;
    }

    return result;
}

static void
wcbe_src_uri_handler_init(gpointer g_iface, gpointer iface_data)
{
    GstURIHandlerInterface *iface = (GstURIHandlerInterface *)g_iface;

    iface->get_type = GST_DEBUG_FUNCPTR(wcbe_src_uri_get_type);
    iface->get_protocols = GST_DEBUG_FUNCPTR(wcbe_src_uri_get_protocols);
    iface->get_uri = GST_DEBUG_FUNCPTR(wcbe_src_uri_get_uri);
    iface->set_uri = GST_DEBUG_FUNCPTR(wcbe_src_uri_set_uri);
}

static GstURIType
wcbe_src_uri_get_type(void)
{
    return GST_URI_SRC;
}

static gchar **
wcbe_src_uri_get_protocols(void)
{
    static gchar *result[] = {"widar", "file", NULL};
    return result;
}

static const gchar *
wcbe_src_uri_get_uri(GstURIHandler *handler)
{
    WcbeSrc *src = WCBE_SRC(handler);
    return src->uri;
}

static gboolean
wcbe_src_uri_set_uri(GstURIHandler *handler, const gchar *uri)
{
    gboolean result = FALSE;
    gboolean wrong_state = FALSE;
    WcbeSrc *src = WCBE_SRC(handler);

    GST_STATE_LOCK(src);
    {
        GstState state = GST_STATE(src);
        if (state != GST_STATE_READY && state != GST_STATE_NULL) {
            wrong_state = TRUE;
        }
    }
    GST_STATE_UNLOCK(src);

    if (!wrong_state) {
        if (src->uri != NULL) {
            g_free(src->uri);
        }
        src->uri = NULL;
        gchar *protocol = gst_uri_get_protocol(uri);
        if (strcmp(protocol, "widar") == 0) {
            gchar *location = gst_uri_get_location(uri);
            gchar *colon = strchr(location, ':');
            if (colon != NULL) {
                gulong port = strtoul(&colon[1], NULL, 10);
                if (port != 0) {
                    *colon = '\0';
                    struct hostent *hostent = gethostbyname(location);
                    if (hostent != NULL && hostent->h_addrtype == AF_INET) {
                        // uri seems to be a hostname or IP address
                        result = TRUE;
                        src->uri = g_strdup(uri);
                    }
                }
            }
            g_free(location);
        }
        else if (strcmp(protocol, "file") == 0) {
            gchar *location = gst_uri_get_location(uri);
            struct stat statbuf;
            int rc = stat(location, &statbuf);
            if (rc != -1 && (statbuf.st_mode & S_IRUSR)) {
                // uri seems to be a filename
                result = TRUE;
                src->uri = g_strdup(uri);
            }
            g_free(location);
        }
        g_free(protocol);
    }
    return result;
}
